from datetime import datetime, UTC

from pandas import DataFrame

from bb_integrations_lib.protocols.pipelines import Step
from bb_integrations_lib.shared.model import FileReference, FileType


class DataFrameFlatFileExportStep(Step):
    def __init__(self, file_path: str, file_type: FileType, *args, **kwargs):
        super().__init__(*args, **kwargs)

        file_path = file_path
        if "{date}" in file_path:
            file_path = file_path.replace("{date}", datetime.now(UTC).strftime("%Y%m%d"))

        self.output = FileReference(file_path, file_type)

    def describe(self) -> str:
        return f"Exporting DataFrame to flat file {self.output.file_path}"

    async def execute(self, i: DataFrame) -> FileReference:
        if self.output.file_type == FileType.excel:
            i.to_excel(self.output.file_path, index=False)
        elif self.output.file_type == FileType.csv:
            i.to_csv(self.output.file_path, index=False)
        else:
            raise NotImplementedError(f"Unsupported file type: {self.output.file_type}")
        return self.output



import json
import os
import uuid
from contextlib import contextmanager, asynccontextmanager
from datetime import datetime, UTC
from functools import lru_cache
from typing import Optional, Any, Iterable, Generator
import re
import loguru
import pytz
import yaml
from bson import ObjectId
from bson.raw_bson import RawBSONDocument
from loguru import logger
from pydantic import ValidationError, BaseModel
from pymongo import MongoClient
from pymongo.synchronous.database import Database

from bb_integrations_lib.secrets.credential_models import FTPCredential, AWSCredential, GoogleCredential, IMAPCredential
from bb_integrations_lib.shared.model import CredentialType


class DateTimeEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime):
            return obj.isoformat()
        return super(DateTimeEncoder, self).default(obj)

class CustomJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime):
            return obj.isoformat()  # Convert datetime to ISO 8601 string
        if isinstance(obj, ObjectId):
            return str(obj)
        if isinstance(obj, BaseModel):
            return obj.model_dump(mode="json")
        return super().default(obj)

class ClientConfig(BaseModel):
    client_name: str
    client_url: str
    client_psk: str
    conn_str: str
    account_username: str = 'costco-integration'


def safe_index(file_name: str, sub_str: str, default: int = 1000) -> int:
    """
    :param file_name: file name.
    :param sub_str: sub string.
    :param default: default value.
    :return: integer index of sub_str.
    """
    if sub_str in file_name:
        return file_name.index(sub_str)
    return default


def file_time(file_name: str, sub_str: str, sep='_', strip_trailing: bool = False) -> str:
    """
    :param file_name: file name.
    :param sub_str: sub string.
    :param sep: string separator.
    :param strip_trailing: if True, strips trailing _digits pattern (e.g., _30380357).
    :return: string time of file.
    """
    if sub_str in file_name:
        start = safe_index(file_name=file_name, sub_str=sub_str) + len(sub_str)
        end = file_name.rfind(".")
        result = file_name[start:end]

        if strip_trailing:
            result = re.sub(r'_\d+$', '', result)

        if sep in result:
            return result.strip(sep)
        return result
    raise IndexError(f"SubString: {sub_str} not found on file: {file_name}")



def file_exact_match(file_name: str, sub_str: str, sep='_') -> bool:
    """
    Check if filename follows this pattern:
    - Starts with the exact substring
    - Optionally followed by separator and date
    - Ends with file extension
    - No other content before or after the substring

    Args:
        file_name: The filename to check
        sub_str: The substring to look for
        sep: The separator character (default '_')

    Returns:
        bool: True if the pattern matches, False otherwise
    """
    if not file_name or not sub_str:
        return False
    if not file_name.startswith(sub_str):
        return False
    remainder = file_name[len(sub_str):]
    if not remainder:
        return True
    if remainder.startswith('.'):
        return True
    if remainder.startswith(sep):
        date_part = remainder[1:]
        date_pattern = r'^(\d{4}-\d{2}-\d{2}|\d{2}-\d{2}-\d{4}|\d{8}|\d{4}_\d{2}_\d{2}|' + \
                       r'\d{2}_\d{2}_\d{4}|\d{4}\d{2}\d{2})(\.[a-zA-Z0-9]+)?$'

        return bool(re.match(date_pattern, date_part))

    return False


def check_if_file_greater_than_date(file_name: str, sub_str: str, date_format: str,
                                    min_date: datetime = datetime.now(UTC),
                                    strip_trailing: bool = False,
                                    file_date_tz: pytz.BaseTzInfo | None = None) -> bool:
    """
    Parses a datetime out of a file name and compares it against the min_date (or current UTC, by default).

    :param file_name: File name to parse.
    :param sub_str: Prefix to strip, if found.
    :param date_format: strptime-compatible date format specifier.
    :param min_date: Date to compare the parsed date against.
      The timezone provided in this object will be used for the parsed timezone if it is not datetime aware.
    :param strip_trailing: if True, strips trailing _digits pattern (e.g., _30380357).
    :param file_date_tz: If provided, assumes the parsed file date is in this timezone
      and converts it to UTC for comparison. Use when file dates are in local time.
    :return: True if the date parsed from file_name is newer than the min_date.
    """
    try:
        file_date = file_time(file_name=file_name, sub_str=sub_str, strip_trailing=strip_trailing)
        date = datetime.strptime(file_date, date_format)
        # If file_date_tz is provided, localize naive date to that tz then convert to UTC
        if date.tzinfo is None or date.tzinfo.utcoffset(date) is None:
            if file_date_tz is not None:
                date = file_date_tz.localize(date).astimezone(pytz.UTC)
                # If min_date is naive, assume it's UTC
                if min_date.tzinfo is None:
                    logger.warning("min_date is naive, assuming UTC")
                    min_date = min_date.replace(tzinfo=pytz.UTC)
            else:
                date = date.replace(tzinfo=min_date.tzinfo)
        return date > min_date
    except IndexError as e:
        logger.error(f"Error: {e}")
        return False


def find_file_in_parent_directories(filename: str, max_levels: int = 20,
                                    secrets_folder: Optional[str] = None) -> str | None:
    """
    Searches for a file in the current directory and up to max_levels of parent directories.

    :param filename: The name of the file to search for.
    :param max_levels: The maximum number of parent directory levels to search.
    :param secrets_folder: Optional parameter defining a secrets folder name.
    :return: The full path to the file if found, else None.
    """
    current_dir = os.getcwd()
    for _ in range(max_levels):
        potential_path = os.path.join(current_dir, filename)
        if os.path.exists(potential_path):
            return potential_path
        # Check in a 'secrets' subdirectory of the current directory.
        if secrets_folder:
            potential_secrets_path = os.path.join(current_dir, secrets_folder, filename)
            if os.path.exists(potential_secrets_path):
                return potential_secrets_path
        current_dir = os.path.dirname(current_dir)
    return None


@lru_cache(maxsize=None)
def load_credentials(credential_type: str = CredentialType.ftp,
                     max_levels: int = 5,
                     secrets_folder_name: str = 'secrets') -> FTPCredential | AWSCredential | GoogleCredential | IMAPCredential:
    """
    :param credential_type: credential type.
    :param max_levels: The maximum number of parent directory levels to search for the credentials file.
    :param secrets_folder_name: The name of a secrets folder. Defaults to secrets.
    :return: Dictionary containing the credentials.
    raise:
            - ValueError if credential type is not supported.
            - FileNotFound if path is not found.
    """
    filename = f'{credential_type}.json'
    path = find_file_in_parent_directories(filename, max_levels, secrets_folder_name)

    if not path:
        logger.error(f"Credentials file not found: {filename}")
        raise FileNotFoundError(f"Credentials file not found: {filename}")

    with open(path, 'r') as file:
        json_credentials = json.load(file)

    match credential_type:
        case CredentialType.ftp:
            return FTPCredential(**json_credentials)
        case CredentialType.aws:
            return AWSCredential(**json_credentials)
        case CredentialType.google:
            return GoogleCredential(**json_credentials)
        case CredentialType.imap:
            return IMAPCredential(**json_credentials)
        case _:
            for CredentialModel in (FTPCredential, AWSCredential, GoogleCredential, IMAPCredential):
                try:
                    return CredentialModel(**json_credentials)
                except (TypeError, ValidationError):
                    continue
            raise TypeError(f'Unable to open: {filename}')


def get_client_config(base_dir: str, config_directory: str = 'deployment_configs',
                      client_name: str = 'coleman') -> ClientConfig:
    try:
        dirs = os.listdir(f'{base_dir}/{config_directory}')
        client_dir = [f for f in dirs if f == client_name][0]
        file_path = f"{base_dir}/{config_directory}/{client_dir}/env-cm.yaml"
        with open(file_path, 'r') as file:
            data = yaml.safe_load(file)
        file_data = data['data']
        return ClientConfig(
            client_name=client_name,
            client_url=file_data['BASE_URL'],
            client_psk=file_data['SYSTEM_PSK'],
            conn_str=build_conn_str(file_data['DB_CONNECT_STR']),
        )
    except IndexError:
        loguru.logger.error(f'Unable to find config file for {client_name}')


def build_conn_str(conn_str) -> str:
    """
    Method to build mongo conn strings w/o pri safely
    :param conn_str: original conn string
    :return: formatted conn string
    """
    if 'localhost' in conn_str:
        return conn_str
    # Determine start of cluster name
    start_cluster = safe_index(file_name=conn_str, sub_str="@") + 1
    # Determine end of cluster name
    end_cluster = safe_index(file_name=conn_str, sub_str='pri') - 1
    # Determine start of connection string
    start = 0
    # Determine end of connection string
    end = safe_index(file_name=conn_str, sub_str="@")
    cluster = conn_str[start_cluster:end_cluster]
    left = conn_str[start:end]
    if 'bbdev' in conn_str:
        return f"{left}@{cluster}.4f2iw.gcp.mongodb.net/"
    return f"{left}@{cluster}.z7gyv.mongodb.net/"


def nested_lookup(iterable: Iterable, key_path: str):
    def get_nested_value(item, path):
        parts = path.split('.')
        value = item
        for part in parts:
            if isinstance(value, dict):
                value = value.get(part)
            else:
                value = getattr(value, part, None)
            if value is None:
                return None
        return value

    return {get_nested_value(i, key_path): i for i in iterable}


def lookup(iterable: Iterable, key: callable):
    return {key(i): i for i in iterable}


@contextmanager
def init_db(connection_str: str, db_name: str):
    from pymongo import MongoClient
    client = MongoClient(connection_str)
    db = client[db_name]
    yield db


@contextmanager
def init_db_async(connection_str: str, db_name: str):
    from pymongo import AsyncMongoClient
    client = AsyncMongoClient(connection_str)
    db = client[db_name]
    try:
        yield db
    finally:
        client.close()


def mongo_client(
        connection_string,
        read_preference='primaryPreferred',
        server_timeout_ms=60000,
        socket_timeout_ms=30000,
        connect_timeout_ms=30000,
        **kwargs
):
    """
    Args:
        connection_string: MongoDB connection URI
        read_preference: 'primary', 'primaryPreferred', 'secondary', 'secondaryPreferred', 'nearest'
        server_timeout_ms: Server selection timeout
        socket_timeout_ms: Socket timeout
        connect_timeout_ms: Connection timeout
        **kwargs: Any other MongoClient options
    """
    return MongoClient(
        connection_string,
        serverSelectionTimeoutMS=server_timeout_ms,
        socketTimeoutMS=socket_timeout_ms,
        connectTimeoutMS=connect_timeout_ms,
        readPreference=read_preference,
        retryWrites=True,
        retryReads=True,
        **kwargs
    )


def gen_lookup(db: Database, collection_name: str, find_params: dict = None, as_raw: bool = False
               ) -> dict[str, dict | RawBSONDocument]:
    """
    Generate a lookup table mapping the _id field of each document in the MongoDB collection to its data.
    :param db: An already-connected pymongo database.
    :param collection_name: The name of the collection.
    :param find_params: Optional parameters to filter the collection on, in Mongo query format
        (will be passed to collection.find).
    :param as_raw: Return RawBSONDocuments, which are decompressed on-the-fly to save resources, instead of plain dicts.
    :return: A dict where the key is the _id field of each document and the value is the whole document.
    """
    collection = db[collection_name]
    if as_raw:
        collection = collection.with_options(
            codec_options=collection.codec_options.with_options(document_class=RawBSONDocument))
    data = list(collection.find(find_params or {}))
    return {str(o['_id']): o for o in data}

class MongoJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, ObjectId):
            return str(obj)
        return super().default(obj)


def is_uuid(s: str):
    """Returns True if string is a valid UUID."""
    try:
        uuid.UUID(s)
        return True
    except ValueError:
        return False


def is_valid_goid(goid: str, prefix: str):
    """Returns true if the string is a valid GOID with the specified prefix."""
    if (not goid.startswith(prefix) or not goid.split(":")[-1].isnumeric()) and not is_uuid(goid):
        return False
    return True





if __name__ == "__main__":
    credentials = load_credentials("google.credentials")
    print(credentials)



import json
from datetime import datetime, timedelta, UTC

from pymongo.asynchronous.database import AsyncDatabase

from bb_integrations_lib.protocols.pipelines import Step
from bb_integrations_lib.util.utils import CustomJSONEncoder


class GetModelHistoryStep(Step):
    def __init__(self, mongo_database: AsyncDatabase, hours_back: float | None = None, include_model_mode: str = "latest_only", *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.mongo_database = mongo_database
        self.hours_back = hours_back
        self.include_model_mode = include_model_mode

    def describe(self) -> str:
        return f"Get model history"

    async def execute(self, i: None) -> str | list:
        if self.include_model_mode == "latest_only":
            return await self.get_latest_model_id()
        else:
            if self.hours_back is not None:
                return await self.get_last_n_models(int(self.hours_back))
            else:
                raise NotImplementedError("Please specify n_hours_back in the configuration")

    async def get_latest_model_id(self) -> str:
        collection = self.mongo_database["model_history"]
        latest_model = await collection.find_one(
            {'status': 'Success'},
            sort=[('time_ran', -1)]
        )
        latest_model_id = str(latest_model['_id'])
        self.pipeline_context.included_files[f'{self.__class__.__name__} result'] = json.dumps(latest_model, cls=CustomJSONEncoder)
        self.pipeline_context.extra_data["latest_model_id"] = latest_model_id
        return latest_model


    async def get_last_n_models(self, n: int) -> list:
        n_hours_ago = datetime.now(UTC) - timedelta(hours=n)
        collection = self.mongo_database["model_history"]
        models = collection.find(
            {'status': 'Success', 'time_ran': {'$gte': n_hours_ago}},
            sort=[('time_ran', -1)]
        )
        models = await models.to_list()
        self.pipeline_context.included_files[f'{self.__class__.__name__} result'] = json.dumps(models, cls=CustomJSONEncoder)
        return models



# This module previously contained PricingIntegrationGetter which has been removed.
# Client creation is now handled via APIFactory and clients are passed directly to PriceMapper.

raise DeprecationWarning("PricingIntegrationGetter has been removed.", )



from typing import Tuple, TypeVar, Generic, Type, Optional

from bb_integrations_lib.gravitate.rita_api import GravitateRitaAPI
from bb_integrations_lib.models.rita.config import GenericConfig
from bb_integrations_lib.secrets import SecretProvider
from bb_integrations_lib.shared.exceptions import ConfigNotFoundError, ConfigValidationError

T = TypeVar("T")




class PipelineWrapper(Generic[T]):
    def __init__(self,
                 bucket_name: str,
                 config_class: Type[T],
                 secret_provider: SecretProvider
                 ):
        self.bucket_name = bucket_name
        self.config_class = config_class
        self.config_name: Optional[str] = None
        self.secret_provider = secret_provider

    async def load_config(self,
                          config_name: str,
                          rita_client: GravitateRitaAPI,
                          ) -> Tuple[T, str, str]:
        """
        Load and validate configuration from Rita API.

        Returns:
            Tuple of (parsed_config, config_id, config_name)

        Raises:
            ConfigNotFoundError: If the configuration is not found
            ConfigValidationError: If the configuration fails validation
        """
        try:
            configs = await rita_client.get_config_by_name(
                bucket_path=self.bucket_name,
                config_name=config_name
            )
        except Exception as e:
            msg = f"Failed to retrieve config '{config_name}' from bucket '{self.bucket_name}': {e}"
            raise ConfigNotFoundError(
                msg
            ) from e

        job_config: GenericConfig = configs[config_name]

        try:
            pipeline_config: T = self.config_class.model_validate(job_config.config)
        except Exception as e:
            raise ConfigValidationError(
                f"Failed to validate config '{config_name}' as {self.config_class.__name__}: {e}"
            ) from e

        return pipeline_config, job_config.config_id, config_name



from pydantic_xml import BaseXmlModel, element


class Sender(
    BaseXmlModel,
):
    load_id: str = element("LoadID")

class Requestor(
    BaseXmlModel,
):
    mpid: str = element("MPID")
    load_id: str = element("LoadID")

class AllowAccessFrom(
    BaseXmlModel,
):
    mpid: str = element("MPID")

class Coordinates(
    BaseXmlModel,
):
    latitude: float = element("Latitude")
    longitude: float = element("Longitude")

class Location(
    BaseXmlModel,
):
    coordinates: Coordinates = element("Coordinates")
    created_date_time: str = element("CreatedDateTime")

class LocationUpdateRequest(
    BaseXmlModel,
    tag="TMSLocationData",
    nsmap={"": 'http://macropoint-lite.com/xml/1.0'}
):
    sender: Sender = element("Sender")
    requestor: Requestor = element("Requestor")
    allow_access_from: AllowAccessFrom = element("AllowAccessFrom")
    location: Location = element("Location")



from typing import Callable

import pandas as pd
import polars as pl
from bb_integrations_lib.protocols.pipelines import Step
from bb_integrations_lib.shared.model import RawData


class ExportDataFrameToRawDataStep(Step):
    def __init__(self, pandas_export_function: str, pandas_export_kwargs: dict,
                 file_name: str | Callable[[], str], *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pandas_export_function = pandas_export_function
        self.pandas_export_kwargs = pandas_export_kwargs
        self.file_name = file_name

    def _get_file_name(self) -> str:
        if callable(self.file_name):
            return self.file_name()
        else:
            return self.file_name

    def describe(self) -> str:
        return "Export a DataFrame to a file wrapped in a RawData object"

    async def execute(self, i: pd.DataFrame | pl.DataFrame) -> RawData:
        func = getattr(i, self.pandas_export_function)
        return RawData(data=func(**self.pandas_export_kwargs).encode("utf-8"), file_name=self._get_file_name())



from bb_integrations_lib.gravitate.sd_api import GravitateSDAPI
from bb_integrations_lib.pipelines.shared.allocation_matcher.matching_utils import (
    MatchedAllocation,
    match_allocations,
)
from bb_integrations_lib.pipelines.shared.allocation_matcher.model import OrderData
from bb_integrations_lib.shared.model import GetOrderBolsAndDropsRequest
from bb_integrations_lib.util.utils import lookup
from loguru import logger
import asyncio


class AllocationMatcher:
    """A matcher for matching executed details to planned details.

    Matches allocated BOLs to their planned details. Carriers generally need
    this information to pay their drivers and bill their customers.

    Can be used to match any list of order numbers. The core matching logic
    is in :func:`match_allocations`, which can be imported and used directly.

    Example:
        Using the class method (recommended)::

            from bb_integrations_lib.pipelines.shared.allocation_matcher.matcher import AllocationMatcher

            results = await AllocationMatcher.match_order_numbers(
                order_numbers=[1103, 1196],
                sd_client=sd_client,
            )

        Using the instance method::

            matcher = AllocationMatcher(order_numbers=[1103, 1196])
            results = await matcher.run_allocation_matching(sd_client)

        Using ``match_allocations`` directly (for custom data sources)::

            from bb_integrations_lib.pipelines.shared.allocation_matcher.matching_utils import match_allocations

            results = match_allocations(
                order_number="1103",
                allocated_bols=bols_and_drops["allocated_bols"],
                executed_bols=bols_and_drops["bols"],
                executed_drops=bols_and_drops["drops"],
                planned_loads=planned_order["loads"],
                planned_drops=planned_order["drops"],
            )
    """
    _semaphore: asyncio.Semaphore | None = None

    def __init__(
            self,
            order_numbers: list[int],
            max_concurrent: int = 10,
    ):
        self.order_numbers = order_numbers
        self.max_concurrent = max_concurrent

    @property
    def semaphore(self) -> asyncio.Semaphore:
        """Lazy-initialized semaphore for rate limiting API calls."""
        if self._semaphore is None:
            self._semaphore = asyncio.Semaphore(self.max_concurrent)
        return self._semaphore

    async def get_planned_details(
            self,
            client: GravitateSDAPI,
            order_number: str,
    ) -> tuple[str, dict | None, str | None]:
        """Get planned order details for a single order number."""
        async with self.semaphore:
            try:
                resp = await client.get_orders(order_number=order_number)
                data = resp.json()
                return order_number, data[0] if data else None, None
            except Exception as e:
                logger.warning(f"Failed to get planned order {order_number}: {e}")
                return order_number, None, str(e)

    async def get_order_data(
            self,
            client: GravitateSDAPI,
    ) -> list[OrderData]:
        """
        Get all planned + executed order data for a list of order numbers.
        """
        results = []

        bols_and_drops_req = GetOrderBolsAndDropsRequest(order_numbers=self.order_numbers)
        bols_response = await client.get_bols_and_drops(bols_and_drops_req)
        all_bols = bols_response.json()
        bols_by_order = lookup(all_bols, lambda x: x['order_number'])
        order_strs = [str(n) for n in self.order_numbers]
        planned_results = await asyncio.gather(
            *[self.get_planned_details(client, n) for n in order_strs]
        )

        for order_num, planned, error in planned_results:
            results.append(OrderData(
                order_number=order_num,
                bols_and_drops=bols_by_order.get(order_num),
                planned_order=planned,
                error=error,
            ))
        return results

    async def run_allocation_matching(
            self,
            sd_client: GravitateSDAPI,
    ) -> list[MatchedAllocation]:
        """
        Get order data and match allocations for all order numbers.
        """
        order_data_list = await self.get_order_data(sd_client)
        all_results: list[MatchedAllocation] = []
        for data in order_data_list:
            if data.error or not data.bols_and_drops or not data.planned_order:
                continue

            results = match_allocations(
                order_number=data.order_number,
                allocated_bols=data.bols_and_drops["allocated_bols"],
                executed_bols=data.bols_and_drops["bols"],
                executed_drops=data.bols_and_drops["drops"],
                planned_loads=data.planned_order["loads"],
                planned_drops=data.planned_order["drops"],
            )
            all_results.extend(results)

        return all_results

    @classmethod
    async def match_order_numbers(
            cls,
            order_numbers: list[int],
            sd_client: GravitateSDAPI,
            max_concurrent: int = 10,
    ):
        return await cls(order_numbers, max_concurrent).run_allocation_matching(sd_client)



from datetime import datetime
from enum import Enum
from typing import Optional

# from pydantic import BaseModel
from pydantic_xml import BaseXmlModel, element

default_serialization_options = {
    "skip_empty": False, "exclude_none": False, "exclude_unset": True
}
"""To be passed to pydantic_xml.BaseXmlModel.to_xml when sending to the KeyVu API"""


class DeliveryStatus(str, Enum):
    planned = "Planned"
    on_route_not_loaded = "OnRouteNotLoaded"
    loading = "Loading"
    on_route_loaded = "OnRouteLoaded"
    unloading = "Unloading"
    delivered = "Delivered"
    confirmed = "Confirmed"
    canceled = "Canceled"


class UnitSystem(str, Enum):
    metric = "Metric"
    imperial = "Imperial"


class KeyVuDeliveryPlan(BaseXmlModel, tag="Loads", nsmap={"": "http://keyvu.com/schemas/carrierloads/v6",
                                                          "xsi": "http://www.w3.org/2001/XMLSchema-instance",
                                                          "xsd": "http://www.w3.org/2001/XMLSchema"}):
    start_date: datetime = element(tag="StartDate")
    end_date: datetime = element(tag="EndDate")
    export_date: datetime = element(tag="ExportDate")
    deliveries: Optional[list["Delivery"]] = element(tag="Delivery", default=None, nillable=True)

class GeoLocation(BaseXmlModel):
    longitude: Optional[float] = element(tag="Longitude", nillable=True, default=None)
    latitude: Optional[float] = element(tag="Latitude", nillable=True, default=None)
    heading: Optional[float] = element(tag="Heading", nillable=True, default=None)
    last_updated: Optional[datetime] = element(tag="LastUpdated", nillable=True, default=None)


class Delivery(BaseXmlModel):
    id: str = element(tag="Id")
    carrier_name: Optional[str] = element(tag="CarrierName", default=None)
    scac: Optional[str] = element(tag="Scac", default=None)
    station_deliveries_array: "StationDeliveryArray" = element(tag="StationDeliveries", default=None)
    trailer: Optional[str] = element(tag="Trailer", default=None)
    last_updated: datetime = element(tag="LastUpdated")
    geo_location: "GeoLocation" = element(tag="GeoLocation", default_factory=lambda: GeoLocation(longitude=None, latitude=None, heading=None, last_updated=None))
    unit: str = element(tag="Unit")

    def __init__(self, station_deliveries: list["StationDelivery"], **kwargs):
        super().__init__(**kwargs)
        self.station_deliveries = station_deliveries

    @property
    def station_deliveries(self) -> list["StationDelivery"] | None:
        return self.station_deliveries_array.station_deliveries

    @station_deliveries.setter
    def station_deliveries(self, new_deliveries: list["StationDelivery"]) -> None:
        self.station_deliveries_array = StationDeliveryArray(station_deliveries=new_deliveries)


class StationDeliveryArray(BaseXmlModel):
    station_deliveries: Optional[list["StationDelivery"]] = element(tag="StationDelivery", default=None)


class StationDelivery(BaseXmlModel):
    delivery_status: DeliveryStatus = element(tag="DeliveryStatus")
    site_id: Optional[str] = element(tag="SiteId", default=None)
    details_array: Optional["StationDeliveryDetailsArray"] = element(tag="Details", default=None)
    bill_of_ladings_array: Optional["StationDeliveryBOLArray"] = element(tag="BillOfLadings", default=None)
    tractor: Optional[str] = element(tag="Tractor", default=None)
    delivery_date: datetime = element(tag="DeliveryDate")
    carrier_delivery_status: Optional[str] = element(tag="CarrierDeliveryStatus", default=None)

    def __init__(self, details: list["StationDeliveryDetails"] | None = None,
                 bill_of_ladings: list["StationDeliveryBOL"] | None = None, **kwargs):
        super().__init__(**kwargs)
        self.details = details
        self.bill_of_ladings = bill_of_ladings

    @property
    def details(self) -> list["StationDeliveryDetails"] | None:
        if not self.details_array:
            return None
        return self.details_array.details

    @details.setter
    def details(self, new_details: list["StationDeliveryDetails"]) -> None:
        self.details_array = StationDeliveryDetailsArray(details=new_details)

    @property
    def bill_of_ladings(self) -> list["StationDeliveryBOL"] | None:
        if not self.bill_of_ladings_array:
            return None
        return self.bill_of_ladings_array.bill_of_ladings

    @bill_of_ladings.setter
    def bill_of_ladings(self, new_bols: list["StationDeliveryBOL"]) -> None:
        self.bill_of_ladings_array = StationDeliveryBOLArray(bill_of_ladings=new_bols)


class StationDeliveryDetailsArray(BaseXmlModel):
    details: Optional[list["StationDeliveryDetails"]] = element(tag="StationDeliveryDetails", default=None)


class StationDeliveryDetails(BaseXmlModel, tag="StationDeliveryDetails", skip_empty=True):
    delivered_volume: Optional[float] = element(tag="DeliveredVolume", nillable=True)
    volume_unit: UnitSystem = element(tag="VolumeUnit")
    tank_sequence: Optional[int] = element(tag="TankSequence", default=None)
    product_id: Optional[str] = element(tag="ProductId", default=None)
    product: Optional[str] = element(tag="Product", default=None)
    is_retained: bool = element(tag="IsRetained")

    @staticmethod
    def from_v1_order_dict(details: dict) -> "StationDeliveryDetails":
        # Prefer using tank sequence, but fall back to product ID based matching. This is the behavior KeyVu prefers.
        tank_sequence = details.get("tank_id")
        product_id = None
        product = None
        if not tank_sequence:
            product_id = details.get("product_source_id")
            product = details.get("product_name")

        return StationDeliveryDetails(
            delivered_volume=details.get("quantity", 0.0),
            volume_unit=UnitSystem.imperial,
            tank_sequence=tank_sequence,
            product_id=product_id,
            product=product,
            is_retained=False # TODO: Support retains
        )


class StationDeliveryBOLArray(BaseXmlModel):
    bill_of_ladings: Optional[list["StationDeliveryBOL"]] = element(tag="StationDeliveryBillOfLadings")


class StationDeliveryBOL(BaseXmlModel):
    terminal_name: Optional[str] = element(tag="TerminalName", nillable=True)
    terminal_control_number: str = element(tag="TerminalControlNumber")
    bill_of_lading_number: str = element(tag="BillOfLadingNumber")
    supplier: str = element(tag="Supplier")
    consignee: str = element(tag="Consignee", default="")



from datetime import datetime, date
from typing import Literal

from pydantic import BaseModel, Field


class JobLocation(BaseModel):
    external_id: str
    name: str
    address: str
    latitude: str
    longitude: str
    city: str
    state: str = Field(min_length=2, max_length=2)
    country_code: str = Field(min_length=2, max_length=2)
    postal_code: str | None = None
    postal_splc:  str | None = None
    time_zone: Literal["US/Pacific", "US/Arizona", "US/Mountain", "US/Central", "US/Eastern"] | None = None

class JobTaskAppointment(BaseModel):
    start_time: datetime
    end_time: datetime
    ready_time: datetime
    late_time: datetime | None = None

class JobTask(BaseModel):
    # external_data: JobTaskExternalData
    remarks: list[str]
    fields: dict[str, str]
    id: str
    external_id: str
    order: int = Field(description="Index of the task within the step.")
    type: str
    status: str
    name: str
    completed: bool
    completed_at: datetime | None = None
    assets: str | None = None

class JobStep(BaseModel):
    tasks: list[JobTask]
    order: int = Field(description="Index of the step within the job.")
    type: str
    name: str
    external_id: str
    location_external_id: str
    completed: bool
    completed_at: datetime | None = None
    is_disabled: bool | None = None
    is_bypassable: bool | None = None
    is_bypassed: bool | None = None
    bypass_reason: str | None = None
    is_reorderable: bool | None = None
    reorder_reason: str | None = None
    is_skippable: bool | None = None
    is_skipped: bool | None = None
    skip_reason: str | None = None

class ValueWithUnit(BaseModel):
    value: float
    uom: str

class ShipmentDetails(BaseModel):
    total_distance: ValueWithUnit

class JobDefinition(BaseModel):
    status: str  # TODO: What statuses are possible?
    external_id: str
    locations: list[JobLocation]
    steps: list[JobStep]
    shipment_details: ShipmentDetails

class LoadEntity(BaseModel):
    type: Literal["trailer", "bill_of_lading"]
    value: str

class LoadDefinition(BaseModel):
    start_date: date
    end_date: date
    load: str | None = None
    user_external_id: str
    entities: list[LoadEntity]



# generated by datamodel-codegen:
#   filename:  temp_openapi.json
#   timestamp: 2025-07-09T01:08:11+00:00

from __future__ import annotations

from datetime import datetime
from enum import Enum
from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel, ConfigDict, EmailStr, Field, RootModel, conint, constr


class BolImagesRequest(BaseModel):
    order_ids: Optional[List[str]] = Field([], title='Order Ids')
    order_numbers: Optional[List[str]] = Field([], title='Order Numbers')


class DeliveryTicketById(BaseModel):
    order_id: str = Field(..., title='Order Id')
    store_id: str = Field(..., title='Store Id')


class DeliveryTicketByNumber(BaseModel):
    order_number: int = Field(..., title='Order Number')
    store_number: str = Field(..., title='Store Number')


class DriverCredentialUpsertRequest(BaseModel):
    driver_id: str = Field(..., title='Driver Id')
    credential_id: str = Field(..., title='Credential Id')
    certification_date: Optional[datetime] = Field(None, title='Certification Date')
    expiration_date: Optional[datetime] = Field(None, title='Expiration Date')


class Type(Enum):
    regular = 'regular'
    order_request = 'order_request'
    backhaul = 'backhaul'


class IntegrationFormResponseOverviewReq(BaseModel):
    form_name: Optional[str] = Field(None, title='Form Name')
    from_date: datetime = Field(..., title='From Date')
    to_date: datetime = Field(..., title='To Date')
    market: Optional[str] = Field(None, title='Market')


class OptimalSupplyReportRequest(BaseModel):
    order_numbers: Optional[List[int]] = Field(None, title='Order Numbers')
    order_ids: Optional[List[str]] = Field(None, title='Order Ids')
    order_date_start: Optional[datetime] = Field(None, title='Order Date Start')
    order_date_end: Optional[datetime] = Field(None, title='Order Date End')
    movement_updated_start: Optional[datetime] = Field(
        None, title='Movement Updated Start'
    )
    movement_updated_end: Optional[datetime] = Field(None, title='Movement Updated End')
    last_change_date: Optional[datetime] = Field(None, title='Last Change Date')


class OptimalSupplyReportRow(BaseModel):
    order_number: int = Field(..., title='Order Number')
    order_id: str = Field(..., title='Order Id')
    site_number: str = Field(..., title='Site Number')
    actual_component_product: str = Field(..., title='Actual Component Product')
    actual_component_product_id: str = Field(..., title='Actual Component Product Id')
    optimal_component_product: str = Field(..., title='Optimal Component Product')
    optimal_component_product_id: str = Field(..., title='Optimal Component Product Id')
    actual_terminal: str = Field(..., title='Actual Terminal')
    actual_terminal_id: str = Field(..., title='Actual Terminal Id')
    optimal_terminal: str = Field(..., title='Optimal Terminal')
    optimal_terminal_id: str = Field(..., title='Optimal Terminal Id')
    actual_supplier: str = Field(..., title='Actual Supplier')
    actual_supplier_id: str = Field(..., title='Actual Supplier Id')
    optimal_supplier: str = Field(..., title='Optimal Supplier')
    optimal_supplier_id: str = Field(..., title='Optimal Supplier Id')
    actual_price_type: str = Field(..., title='Actual Price Type')
    optimal_price_type: str = Field(..., title='Optimal Price Type')
    actual_contract: Optional[str] = Field(None, title='Actual Contract')
    optimal_contract: Optional[str] = Field(None, title='Optimal Contract')
    actual_curve_id: str = Field(..., title='Actual Curve Id')
    optimal_curve_id: str = Field(..., title='Optimal Curve Id')
    actual_price_id: str = Field(..., title='Actual Price Id')
    optimal_price_id: str = Field(..., title='Optimal Price Id')
    actual_loaded_miles: float = Field(..., title='Actual Loaded Miles')
    optimal_loaded_miles: float = Field(..., title='Optimal Loaded Miles')
    actual_product_price: float = Field(..., title='Actual Product Price')
    optimal_product_price: float = Field(..., title='Optimal Product Price')
    actual_freight_rate: float = Field(..., title='Actual Freight Rate')
    optimal_freight_rate: float = Field(..., title='Optimal Freight Rate')
    actual_total_price: float = Field(..., title='Actual Total Price')
    optimal_total_price: float = Field(..., title='Optimal Total Price')
    total_price_delta: float = Field(..., title='Total Price Delta')
    actual_volume: float = Field(..., title='Actual Volume')
    optimal_volume: float = Field(..., title='Optimal Volume')
    last_change_date: datetime = Field(..., title='Last Change Date')
    last_updated_date: datetime = Field(..., title='Last Updated Date')
    reason_code: Optional[str] = Field(None, title='Reason Code')


class PydanticObjectId(
    RootModel[constr(pattern=r'^[0-9a-f]{24}$', min_length=24, max_length=24)]
):
    root: constr(pattern=r'^[0-9a-f]{24}$', min_length=24, max_length=24) = Field(
        ..., examples=['5eb7cf5a86d9755df3a6c593']
    )


class RouteUpsertReq(BaseModel):
    origin_name: str = Field(..., title='Origin Name')
    destination_name: str = Field(..., title='Destination Name')
    api_distance_miles: Optional[float] = Field(None, title='Api Distance Miles')
    api_travel_time_seconds: Optional[float] = Field(
        None, title='Api Travel Time Seconds'
    )
    override_travel_time_seconds: Optional[int] = Field(
        None, title='Override Travel Time Seconds'
    )
    override_distance_miles: Optional[float] = Field(
        None, title='Override Distance Miles'
    )
    override_payroll_miles: Optional[int] = Field(None, title='Override Payroll Miles')


class SalesAdjustedDeliveryUpsertReq(BaseModel):
    source: str = Field(..., title='Source')
    store_id: str = Field(..., title='Store Id')
    tank_id: str = Field(..., title='Tank Id')
    product_id: str = Field(..., title='Product Id')
    volume: float = Field(..., title='Volume')
    date: datetime = Field(..., title='Date')


class SurchargeUpdateReq(BaseModel):
    id: PydanticObjectId
    effective_from: datetime = Field(..., title='Effective From')
    effective_to: Optional[datetime] = Field(None, title='Effective To')
    surcharge: Optional[float] = Field(None, title='Surcharge')


class TractorView(BaseModel):
    id: str = Field(..., title='Id')
    tractor_number: str = Field(..., title='Tractor Number')
    depot: Optional[str] = Field(None, title='Depot')
    vin: Optional[str] = Field(None, title='Vin')
    model: Optional[str] = Field(None, title='Model')
    make: Optional[str] = Field(None, title='Make')
    year: Optional[int] = Field(None, title='Year')
    weight: Optional[int] = Field(None, title='Weight')
    next_maintenance: Optional[datetime] = Field(None, title='Next Maintenance')
    next_tractor_maintenance_required: Optional[bool] = Field(
        None, title='Next Tractor Maintenance Required'
    )
    updated_on: Optional[datetime] = Field(None, title='Updated On')
    extra_data: Optional[Dict[str, Any]] = Field(None, title='Extra Data')


class TrailerConfigMatrixView(BaseModel):
    id: str = Field(..., title='Id')
    importance: int = Field(..., title='Importance')
    trailer_config: Optional[str] = Field(None, title='Trailer Config')
    code: Optional[str] = Field(None, title='Code')
    values: Optional[List[float]] = Field([], title='Values')
    updated_on: Optional[datetime] = Field(None, title='Updated On')


class TrailerView(BaseModel):
    id: str = Field(..., title='Id')
    trailer_number: str = Field(..., title='Trailer Number')
    configuration: Optional[str] = Field(None, title='Configuration')
    depot: Optional[str] = Field(None, title='Depot')
    make: Optional[str] = Field(None, title='Make')
    model: Optional[str] = Field(None, title='Model')
    weight: Optional[float] = Field(None, title='Weight')
    updated_on: Optional[datetime] = Field(None, title='Updated On')
    extra_data: Optional[Dict[str, Any]] = Field({}, title='Extra Data')


class UpsertTractorReq(BaseModel):
    id: Optional[str] = Field(None, title='Id')
    tractor_number: Optional[str] = Field(None, title='Tractor Number')
    depot: Optional[str] = Field(None, title='Depot')
    vin: Optional[str] = Field(None, title='Vin')
    make: Optional[str] = Field(None, title='Make')
    model: Optional[str] = Field(None, title='Model')
    year: Optional[int] = Field(None, title='Year')
    weight: Optional[int] = Field(None, title='Weight')
    next_maintenance: Optional[datetime] = Field(None, title='Next Maintenance')
    next_tractor_maintenance_required: Optional[bool] = Field(
        None, title='Next Tractor Maintenance Required'
    )
    source_id: Optional[str] = Field(None, title='Source Id')
    source_system_id: Optional[str] = Field(None, title='Source System Id')


class UpsertTractorResp(BaseModel):
    created: Optional[int] = Field(0, title='Created')
    updated: Optional[int] = Field(0, title='Updated')
    errors: List[str] = Field(..., title='Errors')


class UpsertTrailerReq(BaseModel):
    id: Optional[str] = Field(None, title='Id')
    trailer_number: Optional[str] = Field(None, title='Trailer Number')
    configuration: Optional[str] = Field(None, title='Configuration')
    depot: Optional[str] = Field(None, title='Depot')
    make: Optional[str] = Field(None, title='Make')
    model: Optional[str] = Field(None, title='Model')
    weight: Optional[float] = Field(None, title='Weight')
    source_id: Optional[str] = Field(None, title='Source Id')
    source_system_id: Optional[str] = Field(None, title='Source System Id')


class UpsertTrailerResp(BaseModel):
    created: Optional[int] = Field(0, title='Created')
    updated: Optional[int] = Field(0, title='Updated')
    errors: List[str] = Field(..., title='Errors')


class AllocatedBolResponse(RootModel[Any]):
    root: Any


class BillableDateType(RootModel[Any]):
    root: Any


class BillableVolumeType(RootModel[Any]):
    root: Any


class BlendView(RootModel[Any]):
    root: Any


class BolImageResponse(RootModel[Any]):
    root: Any


class Card(RootModel[Any]):
    root: Any


class ContractVolumeRequest(RootModel[Any]):
    root: Any


class Coordinate(RootModel[Any]):
    root: Any


class Counterparty(RootModel[Any]):
    root: Any


class DateWindow(RootModel[Any]):
    root: Any


class DaysOfWeek(RootModel[Any]):
    root: Any


class DeliveryRequestStatus(RootModel[Any]):
    root: Any


class DeliveryWindow(RootModel[Any]):
    root: Any


class DeliveryWindowDefault(RootModel[Any]):
    root: Any


class DriverCredentialView(RootModel[Any]):
    root: Any


class DriverTerminalCardView(RootModel[Any]):
    root: Any


class DropCreateRequest(RootModel[Any]):
    root: Any


class DropWorkflow(RootModel[Any]):
    root: Any


class Dwell(RootModel[Any]):
    root: Any


class EBolDetailRequest(RootModel[Any]):
    root: Any


class ExternalBOLDetailRequest(RootModel[Any]):
    root: Any


class FormFieldResponse(RootModel[Any]):
    root: Any


class FreightCost(RootModel[Any]):
    root: Any


class FreightItem(RootModel[Any]):
    root: Any


class FreightTransactionOutput(RootModel[Any]):
    root: Any


class FreightTransactionDetailOutput(RootModel[Any]):
    root: Any


class FreightTransactionRowOutput(RootModel[Any]):
    root: Any


class InCabSupplyOptionMode(RootModel[Any]):
    root: Any


class InCabTripMode(RootModel[Any]):
    root: Any


class InNetworkSupplyZone(RootModel[Any]):
    root: Any


class InNetworkTerminal(RootModel[Any]):
    root: Any


class InvoiceDynamicStatus(RootModel[Any]):
    root: Any


class InvoiceStaticStatus(RootModel[Any]):
    root: Any


class InvoiceType(RootModel[Any]):
    root: Any


class Key(RootModel[Any]):
    root: Any


class LocationResponse(RootModel[Any]):
    root: Any


class MonitoringStrategy(RootModel[Any]):
    root: Any


class NOSupplyOptionResponse(RootModel[Any]):
    root: Any


class OrderCreateStatus(RootModel[Any]):
    root: Any


class OrderDriver(RootModel[Any]):
    root: Any


class OrderReqNotificationStates(RootModel[Any]):
    root: Any


class OrderResponse(RootModel[Any]):
    root: Any


class OrderState(RootModel[Any]):
    root: Any


class OrderType(RootModel[Any]):
    root: Any


class PayrollExportDetailModel(RootModel[Any]):
    root: Any


class PriceType(RootModel[Any]):
    root: Any


class ProductGroups(RootModel[Any]):
    root: Any


class ProductIDName(RootModel[Any]):
    root: Any


class RateBookType(RootModel[Any]):
    root: Any


class SaveDropDetail(RootModel[Any]):
    root: Any


class SaveDropMode(RootModel[Any]):
    root: Any


class Shift(RootModel[Any]):
    root: Any


class SourceMap(RootModel[Any]):
    root: Any


class SourcingStrategy(RootModel[Any]):
    root: Any


class StatusUpdate(RootModel[Any]):
    root: Any


class StoreStatus(RootModel[Any]):
    root: Any


class StoreTank(RootModel[Any]):
    root: Any


class SupplyPriceUpdateResponse(RootModel[Any]):
    root: Any


class SurchargeType(RootModel[Any]):
    root: Any


class TankLidEnum(RootModel[Any]):
    root: Any


class TerminalType(RootModel[Any]):
    root: Any


class TimezoneEnum(RootModel[Any]):
    root: Any


class UnavailableHours(RootModel[Any]):
    root: Any


class UnavailableHoursRequest(RootModel[Any]):
    root: Any


class ValidationError(RootModel[Any]):
    root: Any


class VolumeDistributionRequest(RootModel[Any]):
    root: Any


class WaterReadingRequirement(RootModel[Any]):
    root: Any


class ForecastApiCreateRequest(RootModel[Any]):
    root: Any


class MarketSchemaSectorView(RootModel[Any]):
    root: Any


class OrderSchemaBolResponse(RootModel[Any]):
    root: Any


class OrderSchemaDropResponse(RootModel[Any]):
    root: Any


class BolAndDropResponse(BaseModel):
    order_number: Optional[str] = Field(None, title='Order Number')
    order_id: Optional[str] = Field(None, title='Order Id')
    po: Optional[str] = Field(None, title='Po')
    carrier_id: Optional[str] = Field(None, title='Carrier Id')
    last_movement_update: Optional[datetime] = Field(None, title='Last Movement Update')
    order_date: Optional[datetime] = Field(None, title='Order Date')
    status: Optional[str] = Field(None, title='Status')
    type: Optional[str] = Field(None, title='Type')
    drops: Optional[List[OrderSchemaDropResponse]] = Field(None, title='Drops')
    bols: Optional[List[OrderSchemaBolResponse]] = Field(None, title='Bols')
    costs: Optional[List[FreightCost]] = Field(None, title='Costs')
    validation_bypass_on: Optional[datetime] = Field(None, title='Validation Bypass On')
    has_additives: Optional[bool] = Field(False, title='Has Additives')
    estimated_freight: Optional[FreightTransactionOutput] = None
    actual_freight: Optional[FreightTransactionOutput] = None
    allocated_bol_error: Optional[str] = Field(None, title='Allocated Bol Error')
    allocated_bol_issue: Optional[str] = Field(None, title='Allocated Bol Issue')
    allocated_bols: Optional[List[AllocatedBolResponse]] = Field(
        None, title='Allocated Bols'
    )
    last_change_date: Optional[datetime] = Field(None, title='Last Change Date')
    reference_order_number: Optional[str] = Field(None, title='Reference Order Number')


class BolImagesResponse(BaseModel):
    order_id: Optional[str] = Field(None, title='Order Id')
    bol_id: Optional[str] = Field(None, title='Bol Id')
    bol_number: Optional[str] = Field(None, title='Bol Number')
    photos: Optional[List[BolImageResponse]] = Field(None, title='Photos')


class BolsAndDropsRequest(BaseModel):
    order_date_start: Optional[datetime] = Field(None, title='Order Date Start')
    order_date_end: Optional[datetime] = Field(None, title='Order Date End')
    movement_updated_start: Optional[datetime] = Field(
        None, title='Movement Updated Start'
    )
    movement_updated_end: Optional[datetime] = Field(None, title='Movement Updated End')
    order_ids: Optional[List[str]] = Field([], title='Order Ids')
    order_numbers: Optional[List[int]] = Field([], title='Order Numbers')
    order_states: Optional[List[OrderState]] = Field([], title='Order States')
    order_types: Optional[List[OrderType]] = Field([], title='Order Types')
    include_invalid: Optional[bool] = Field(False, title='Include Invalid')
    include_bol_allocation: Optional[bool] = Field(
        False, title='Include Bol Allocation'
    )
    last_change_date: Optional[datetime] = Field(None, title='Last Change Date')
    reference_order_numbers: Optional[List[str]] = Field(
        [], title='Reference Order Numbers'
    )


class CounterpartyView(BaseModel):
    id: str = Field(..., title='Id')
    name: str = Field(..., title='Name')
    goid: Optional[str] = Field(None, title='Goid')
    scac: Optional[str] = Field(None, title='Scac')
    types: Optional[List[str]] = Field(None, title='Types')
    carrier_type: Optional[str] = Field(None, title='Carrier Type')
    trailer_config: Optional[str] = Field(None, title='Trailer Config')
    source_id: Optional[str] = Field(None, title='Source Id')
    source_system: Optional[str] = Field(None, title='Source System')
    source_system_id: Optional[str] = Field(None, title='Source System Id')
    emails: Optional[List[str]] = Field(None, title='Emails')
    sourcing_strategy: Optional[SourcingStrategy] = None
    extra_data: Optional[Dict[str, Any]] = Field(None, title='Extra Data')
    updated_on: Optional[datetime] = Field(None, title='Updated On')
    allow_short_loads: Optional[bool] = Field(False, title='Allow Short Loads')
    order_notification_preferences: Optional[List[OrderReqNotificationStates]] = Field(
        [], title='Order Notification Preferences'
    )
    supply_map: Optional[List[SourceMap]] = Field([], title='Supply Map')


class CreateGroupRequest(BaseModel):
    source_id: Optional[str] = Field(None, title='Source Id')
    name: str = Field(..., title='Name')
    keys: List[Key] = Field(..., title='Keys')
    as_of: datetime = Field(..., title='As Of')
    min: Optional[float] = Field(None, title='Min')
    max: Optional[float] = Field(None, title='Max')
    start_hour: Optional[int] = Field(None, title='Start Hour')
    volume_distributions: Optional[List[VolumeDistributionRequest]] = Field(
        [], title='Volume Distributions'
    )
    contract_volumes: Optional[List[ContractVolumeRequest]] = Field(
        [], title='Contract Volumes'
    )
    daily_percent: Optional[float] = Field(None, title='Daily Percent')
    weekly_percent: Optional[float] = Field(None, title='Weekly Percent')
    monthly_percent: Optional[float] = Field(None, title='Monthly Percent')
    contract_start: Optional[datetime] = Field(None, title='Contract Start')
    contract_end: Optional[datetime] = Field(None, title='Contract End')
    week_start_day: Optional[DaysOfWeek] = 'monday'


class DriverView(BaseModel):
    id: str = Field(..., title='Id')
    name: Optional[str] = Field('', title='Name')
    username: Optional[str] = Field('', title='Username')
    depot_id: Optional[str] = Field(None, title='Depot Id')
    depot_name: Optional[str] = Field(None, title='Depot Name')
    in_cab_trip_mode: Optional[InCabTripMode] = Field(
        default_factory=lambda: InCabTripMode.model_validate('all')
    )
    in_cab_supply_option_mode: Optional[InCabSupplyOptionMode] = Field(
        default_factory=lambda: InCabSupplyOptionMode.model_validate('disabled')
    )
    trailer_number: Optional[str] = Field(None, title='Trailer Number')
    tractor_number: Optional[str] = Field(None, title='Tractor Number')
    updated_on: Optional[datetime] = Field(None, title='Updated On')
    extra_data: Optional[Dict[str, Any]] = Field(None, title='Extra Data')
    cards: Optional[List[DriverTerminalCardView]] = Field([], title='Cards')
    credentials: Optional[List[DriverCredentialView]] = Field([], title='Credentials')


class EBolRequest(BaseModel):
    source_system: str = Field(..., title='Source System')
    bol_number: str = Field(..., title='Bol Number')
    date: datetime = Field(..., title='Date')
    terminal_id: str = Field(..., title='Terminal Id')
    supplier_id: str = Field(..., title='Supplier Id')
    details: List[EBolDetailRequest] = Field(..., title='Details')


class ExternalBOLRequest(BaseModel):
    order_id: str = Field(..., title='Order Id')
    bol_number: str = Field(..., title='Bol Number')
    terminal_id: str = Field(..., title='Terminal Id')
    bol_date: datetime = Field(..., title='Bol Date')
    details: List[ExternalBOLDetailRequest] = Field(..., title='Details')


class FormResponseRow(BaseModel):
    id: PydanticObjectId
    name: str = Field(..., title='Name')
    description: Optional[str] = Field(None, title='Description')
    order_number: int = Field(..., title='Order Number')
    driver_name: str = Field(..., title='Driver Name')
    date: datetime = Field(..., title='Date')
    shift: str = Field(..., title='Shift')
    location_name: Optional[str] = Field(None, title='Location Name')
    address: Optional[str] = Field(None, title='Address')
    latitude: Optional[float] = Field(None, title='Latitude')
    longitude: Optional[float] = Field(None, title='Longitude')
    market: Optional[str] = Field(None, title='Market')
    sector: Optional[str] = Field(None, title='Sector')
    supply_zones: Optional[List[str]] = Field([], title='Supply Zones')
    tractor: Optional[str] = Field(None, title='Tractor')
    trailer_number: Optional[str] = Field(None, title='Trailer Number')
    required: Optional[bool] = Field(False, title='Required')
    response: List[FormFieldResponse] = Field(..., title='Response')


class FreightIntegrationReq(BaseModel):
    order_numbers: Optional[List[int]] = Field(None, title='Order Numbers')
    order_ids: Optional[List[PydanticObjectId]] = Field(None, title='Order Ids')
    last_change_date: Optional[datetime] = Field(None, title='Last Change Date')


class GetOrderRequestResponse(BaseModel):
    type: Optional[str] = Field(None, title='Type')
    number: int = Field(..., title='Number')
    order_id: Optional[str] = Field(None, title='Order Id')
    order_number: Optional[int] = Field(None, title='Order Number')
    order_date: Optional[datetime] = Field(None, title='Order Date')
    order_state: Optional[Union[DeliveryRequestStatus, str]] = Field(
        None, title='Order State'
    )
    reference_order_number: Optional[str] = Field(None, title='Reference Order Number')
    last_change_date: Optional[datetime] = Field(None, title='Last Change Date')
    site_name: Optional[str] = Field(None, title='Site Name')
    site_id: Optional[str] = Field(None, title='Site Id')
    customer_id: Optional[str] = Field(None, title='Customer Id')
    customer_name: Optional[str] = Field(None, title='Customer Name')
    delivery_window_start: datetime = Field(..., title='Delivery Window Start')
    delivery_window_end: datetime = Field(..., title='Delivery Window End')
    products: List[ProductIDName] = Field(..., title='Products')
    extra_data: Optional[Dict[str, Any]] = Field({}, title='Extra Data')
    status: str = Field(..., title='Status')


class GetOrderResponse(BaseModel):
    type: Optional[Type] = Field('regular', title='Type')
    order_id: Optional[str] = Field(None, title='Order Id')
    drivers: Optional[List[OrderDriver]] = Field([], title='Drivers')
    order_number: Optional[int] = Field(None, title='Order Number')
    order_date: Optional[datetime] = Field(None, title='Order Date')
    order_state: Optional[str] = Field(None, title='Order State')
    carrier_window_start: Optional[datetime] = Field(None, title='Carrier Window Start')
    carrier_window_end: Optional[datetime] = Field(None, title='Carrier Window End')
    carrier_notify_state: Optional[str] = Field(None, title='Carrier Notify State')
    load_window_start: Optional[datetime] = Field(None, title='Load Window Start')
    load_window_end: Optional[datetime] = Field(None, title='Load Window End')
    dispatch_window_start: Optional[datetime] = Field(
        None, title='Dispatch Window Start'
    )
    dispatch_window_end: Optional[datetime] = Field(None, title='Dispatch Window End')
    hauler_counterparty_id: Optional[str] = Field(None, title='Hauler Counterparty Id')
    hauler_counterparty_name: Optional[str] = Field(
        None, title='Hauler Counterparty Name'
    )
    hauler_counterparty_source_id: Optional[str] = Field(
        None, title='Hauler Counterparty Source Id'
    )
    hauler_counterparty_source_system: Optional[str] = Field(
        None, title='Hauler Counterparty Source System'
    )
    hauled_by_updated_by: Optional[str] = Field(None, title='Hauled By Updated By')
    hauled_by_updated: Optional[datetime] = Field(None, title='Hauled By Updated')
    loads: Optional[List[LocationResponse]] = Field([], title='Loads')
    drops: Optional[List[LocationResponse]] = Field([], title='Drops')
    trip_status: Optional[str] = Field(None, title='Trip Status')
    last_change_date: Optional[datetime] = Field(None, title='Last Change Date')
    market: Optional[str] = Field(None, title='Market')
    supply_option: Optional[NOSupplyOptionResponse] = None
    created_by: Optional[str] = Field(None, title='Created By')
    note: Optional[str] = Field(None, title='Note')
    estimated_load_minutes: Optional[int] = Field(None, title='Estimated Load Minutes')
    total_miles: Optional[float] = Field(None, title='Total Miles')
    loaded_miles: Optional[float] = Field(None, title='Loaded Miles')
    unloaded_miles: Optional[float] = Field(None, title='Unloaded Miles')
    reference_order_number: Optional[str] = Field(None, title='Reference Order Number')
    extra_data: Optional[Dict[str, Any]] = Field({}, title='Extra Data')


class GetOrdersRequest(BaseModel):
    order_id: Optional[str] = Field(None, title='Order Id')
    order_number: Optional[int] = Field(None, title='Order Number')
    type: Optional[str] = Field(None, title='Type')
    state: Optional[Union[OrderState, str]] = Field(None, title='State')
    last_change_date: Optional[datetime] = Field(None, title='Last Change Date')
    order_date_start: Optional[datetime] = Field(None, title='Order Date Start')
    order_date_end: Optional[datetime] = Field(None, title='Order Date End')
    reference_order_number: Optional[str] = Field(None, title='Reference Order Number')
    order_date: Optional[datetime] = Field(None, title='Order Date')


class HTTPValidationError(BaseModel):
    detail: Optional[List[ValidationError]] = Field(None, title='Detail')


class ImportRequest(BaseModel):
    reqs: List[ForecastApiCreateRequest] = Field(..., max_length=5000, title='Reqs')


class InvoiceAllReq(BaseModel):
    order_numbers: Optional[List[int]] = Field(None, title='Order Numbers')
    order_ids: Optional[List[str]] = Field(None, title='Order Ids')
    invoice_numbers: Optional[List[str]] = Field(None, title='Invoice Numbers')
    status: Optional[InvoiceStaticStatus] = 'sent'
    counterparty_name: Optional[str] = Field(None, title='Counterparty Name')
    counterparty_id: Optional[str] = Field(None, title='Counterparty Id')
    book_type: Optional[RateBookType] = Field(
        default_factory=lambda: RateBookType.model_validate('Revenue')
    )
    as_of: Optional[datetime] = Field(None, title='As Of')
    between: Optional[DateWindow] = None


class InvoiceRow(BaseModel):
    model_config = ConfigDict(
        extra='allow',
    )
    invoice_number: str = Field(..., title='Invoice Number')
    sent_date: Optional[datetime] = Field(None, title='Sent Date')
    when_to_send_date: Optional[datetime] = Field(None, title='When To Send Date')
    created_on: Optional[datetime] = Field(None, title='Created On')
    type: InvoiceType
    status: Optional[Union[InvoiceStaticStatus, InvoiceDynamicStatus]] = Field(
        None, title='Status'
    )
    transactions: List[FreightTransactionRowOutput] = Field(..., title='Transactions')
    note: Optional[str] = Field(None, title='Note')
    base_amount: float = Field(..., title='Base Amount')
    base_distance: float = Field(..., title='Base Distance')
    distance_uom: str = Field(..., title='Distance Uom')
    accessorial_amount: float = Field(..., title='Accessorial Amount')
    surcharge_amount: float = Field(..., title='Surcharge Amount')
    total_amount: float = Field(..., title='Total Amount')


class LocationView(BaseModel):
    model_config = ConfigDict(
        extra='allow',
    )
    id: str = Field(..., title='Id')
    name: str = Field(..., title='Name')
    short_name: Optional[str] = Field(None, title='Short Name')
    market: Optional[str] = Field(None, title='Market')
    market_id: Optional[str] = Field(None, title='Market Id')
    freight_region_name: Optional[str] = Field(None, title='Freight Region Name')
    freight_region_id: Optional[str] = Field(None, title='Freight Region Id')
    type: Optional[str] = Field(None, title='Type')
    lat: Optional[float] = Field(None, title='Lat')
    lon: Optional[float] = Field(None, title='Lon')
    address: Optional[str] = Field(None, title='Address')
    phone: Optional[str] = Field(None, title='Phone')
    city: Optional[str] = Field(None, title='City')
    state: Optional[str] = Field(None, title='State')
    active: Optional[bool] = Field(None, title='Active')
    postal_code: Optional[str] = Field(None, title='Postal Code')
    source_id: Optional[str] = Field(None, title='Source Id')
    source_system: Optional[str] = Field(None, title='Source System')
    source_system_id: Optional[str] = Field(None, title='Source System Id')
    timezone: Optional[str] = Field(None, title='Timezone')
    supply_zones: Optional[List[str]] = Field([], title='Supply Zones')
    updated_on: Optional[datetime] = Field(None, title='Updated On')
    cards: Optional[List[Card]] = Field([], title='Cards')
    dwells: Optional[List[Dwell]] = Field([], title='Dwells')
    requires_card: Optional[bool] = Field(None, title='Requires Card')
    geofence: Optional[List[Coordinate]] = Field(None, title='Geofence')
    supply_map: Optional[List[SourceMap]] = Field([], title='Supply Map')
    extra_data: Optional[Dict[str, Any]] = Field({}, title='Extra Data')


class MarketView(BaseModel):
    id: str = Field(..., title='Id')
    name: str = Field(..., title='Name')
    network_radius: Optional[int] = Field(250, title='Network Radius')
    active: Optional[bool] = Field(True, title='Active')
    trailer_config: Optional[str] = Field(None, title='Trailer Config')
    updated_on: Optional[datetime] = Field(None, title='Updated On')
    extra_data: Optional[Dict[str, Any]] = Field(None, title='Extra Data')
    sectors: Optional[List[MarketSchemaSectorView]] = Field([], title='Sectors')
    delivery_window_default: Optional[DeliveryWindowDefault] = {
        'name': 'default',
        'cutoffs': [
            {
                'cutoff_time': 16,
                'earliest_request_hours': 6,
                'earliest_request_offset_days': 1,
            }
        ],
    }


class OrderCreateRequest(BaseModel):
    model_config = ConfigDict(
        extra='allow',
    )
    reference_order_number: Optional[str] = Field(..., title='Reference Order Number')
    supply_owner: Optional[Counterparty] = None
    sourcing_strategy: Optional[SourcingStrategy] = None
    manual_supply_fallback: Optional[bool] = Field(True, title='Manual Supply Fallback')
    allow_alternate_products: Optional[bool] = Field(
        False, title='Allow Alternate Products'
    )
    delivery_window: DeliveryWindow
    fit_to_trailer: Optional[bool] = Field(False, title='Fit To Trailer')
    note: Optional[str] = Field(None, title='Note')
    drops: List[DropCreateRequest] = Field(..., title='Drops')
    accept_by: Optional[datetime] = Field(None, title='Accept By')
    extra_data: Optional[Dict[str, Any]] = Field({}, title='Extra Data')


class OrderCreateResponse(BaseModel):
    status: OrderCreateStatus
    order_number: Optional[int] = Field(None, title='Order Number')
    order: Optional[OrderResponse] = None
    errors: Optional[List[str]] = Field([], title='Errors')
    accept_by: Optional[datetime] = Field(None, title='Accept By')
    reference_order_number: Optional[str] = Field(None, title='Reference Order Number')
    extra_data: Optional[Dict[str, str]] = Field(None, title='Extra Data')


class OrderFreightResp(BaseModel):
    number: int = Field(..., title='Number')
    po: Optional[str] = Field(None, title='Po')
    freight_rate: float = Field(..., title='Freight Rate')
    freight_total: float = Field(..., title='Freight Total')
    freight_items: List[FreightItem] = Field(..., title='Freight Items')


class OrderFreightRespV2(BaseModel):
    number: int = Field(..., title='Number')
    po: Optional[str] = Field(None, title='Po')
    freight_rate: float = Field(..., title='Freight Rate')
    freight_total: float = Field(..., title='Freight Total')
    freight_items: List[FreightTransactionDetailOutput] = Field(
        ..., title='Freight Items'
    )


class OrderStatusUpdateRequest(BaseModel):
    order_id: str = Field(..., title='Order Id')
    order_number: Optional[str] = Field(None, title='Order Number')
    status: StatusUpdate
    location_id: Optional[str] = Field(None, title='Location Id')
    eta: Optional[datetime] = Field(None, title='Eta')
    actual: Optional[datetime] = Field(None, title='Actual')


class PayrollExportDetailResponse(BaseModel):
    driver_name: str = Field(..., title='Driver Name')
    driver_source_id: Optional[str] = Field(None, title='Driver Source Id')
    driver_source_system: str = Field(..., title='Driver Source System')
    end_date: datetime = Field(..., title='End Date')
    hours_worked: float = Field(..., title='Hours Worked')
    pay_earned: float = Field(..., title='Pay Earned')
    payroll_config: Optional[str] = Field(None, title='Payroll Config')
    start_date: datetime = Field(..., title='Start Date')
    status: str = Field(..., title='Status')
    updated: datetime = Field(..., title='Updated')
    detail: Optional[List[PayrollExportDetailModel]] = Field([], title='Detail')


class PriceAllRequest(BaseModel):
    as_of: Optional[datetime] = Field(None, title='As Of')
    last_change_date: Optional[datetime] = Field(None, title='Last Change Date')
    terminals: Optional[List[str]] = Field(None, title='Terminals')
    suppliers: Optional[List[str]] = Field(None, title='Suppliers')
    products: Optional[List[str]] = Field(None, title='Products')
    price_types: Optional[List[PriceType]] = Field(None, title='Price Types')
    product_groups: Optional[List[ProductGroups]] = Field(None, title='Product Groups')


class PriceResponse(BaseModel):
    curve_id: str = Field(..., title='Curve Id')
    price_id: str = Field(..., title='Price Id')
    city: Optional[str] = Field(None, title='City')
    contract: Optional[str] = Field(None, title='Contract')
    product: Optional[str] = Field(None, title='Product')
    product_id: str = Field(..., title='Product Id')
    supplier: Optional[str] = Field(None, title='Supplier')
    supplier_id: str = Field(..., title='Supplier Id')
    counterparty: Optional[str] = Field(None, title='Counterparty')
    counterparty_id: Optional[str] = Field(None, title='Counterparty Id')
    terminal: Optional[str] = Field(None, title='Terminal')
    terminal_id: str = Field(..., title='Terminal Id')
    store_number: Optional[str] = Field(None, title='Store Number')
    store_id: Optional[str] = Field(None, title='Store Id')
    price_type: str = Field(..., title='Price Type')
    product_group: Optional[str] = Field(None, title='Product Group')
    price: float = Field(..., title='Price')
    effective_from: datetime = Field(..., title='Effective From')
    effective_to: datetime = Field(..., title='Effective To')
    expire: Optional[datetime] = Field(None, title='Expire')
    disabled: bool = Field(..., title='Disabled')
    disabled_by: Optional[str] = Field(None, title='Disabled By')
    disabled_until: Optional[datetime] = Field(None, title='Disabled Until')
    disabled_reason: Optional[str] = Field(None, title='Disabled Reason')
    updated_by: Optional[str] = Field(None, title='Updated By')
    updated_on: Optional[datetime] = Field(None, title='Updated On')
    extra_data: Optional[Dict[str, Any]] = Field(None, title='Extra Data')
    group_id: Optional[str] = Field(None, title='Group Id')
    group_effective_cutover: Optional[datetime] = Field(
        None, title='Group Effective Cutover'
    )
    group_effective_cutover_timezone: Optional[str] = Field(
        None, title='Group Effective Cutover Timezone'
    )
    group_identifier: Optional[str] = Field(None, title='Group Identifier')
    group_name: Optional[str] = Field(None, title='Group Name')
    min_constraint: Optional[float] = Field(None, title='Min Constraint')
    max_constraint: Optional[float] = Field(None, title='Max Constraint')
    created_on: Optional[datetime] = Field(None, title='Created On')
    net_or_gross_type: Optional[BillableVolumeType] = None
    contract_lifting_valuation_method: Optional[BillableDateType] = None


class ProductView(BaseModel):
    id: str = Field(..., title='Id')
    name: str = Field(..., title='Name')
    short_name: Optional[str] = Field(None, title='Short Name')
    group: str = Field(..., title='Group')
    weight_group: str = Field(..., title='Weight Group')
    icon: str = Field(..., title='Icon')
    source_id: Optional[str] = Field(None, title='Source Id')
    source_system: Optional[str] = Field(None, title='Source System')
    source_system_id: Optional[str] = Field(None, title='Source System Id')
    extra_data: Optional[Dict[str, Any]] = Field({}, title='Extra Data')
    blends: Optional[List[BlendView]] = Field([], title='Blends')
    alternate_products: Optional[List[str]] = Field([], title='Alternate Products')
    updated_on: Optional[datetime] = Field(None, title='Updated On')
    tank_lid_code_id: Optional[str] = Field(None, title='Tank Lid Code Id')
    supply_map: Optional[List[SourceMap]] = Field([], title='Supply Map')


class SaveDropReq(BaseModel):
    mode: Optional[SaveDropMode] = Field(
        default_factory=lambda: SaveDropMode.model_validate('append')
    )
    order_id: str = Field(..., title='Order Id')
    location_id: str = Field(..., title='Location Id')
    details: List[SaveDropDetail] = Field(..., title='Details')


class StoreV2(BaseModel):
    model_config = ConfigDict(
        extra='allow',
    )
    field_id: Optional[PydanticObjectId] = Field(
        None, alias='_id', description='MongoDB document ObjectID'
    )
    store_number: Optional[str] = Field(None, title='Store Number')
    name: Optional[str] = Field(None, title='Name')
    market: Optional[str] = Field(None, title='Market')
    market_id: Optional[str] = Field(None, title='Market Id')
    sector: Optional[str] = Field(None, title='Sector')
    sector_id: Optional[str] = Field(None, title='Sector Id')
    tanks: Optional[List[StoreTank]] = Field([], title='Tanks')
    lat: Optional[float] = Field(None, title='Lat')
    lon: Optional[float] = Field(None, title='Lon')
    address: Optional[str] = Field(None, title='Address')
    city: Optional[str] = Field(None, title='City')
    state: Optional[str] = Field(None, title='State')
    postal_code: Optional[str] = Field(None, title='Postal Code')
    contact_name: Optional[str] = Field(None, title='Contact Name')
    phone_number: Optional[str] = Field(None, title='Phone Number')
    emails: Optional[List[str]] = Field([], title='Emails')
    hours: Optional[str] = Field('', title='Hours')
    timezone: Optional[str] = Field('US/Central', title='Timezone')
    status: Optional[StoreStatus] = Field(
        default_factory=lambda: StoreStatus.model_validate('open')
    )
    drop_workflow: Optional[DropWorkflow] = Field(
        default_factory=lambda: DropWorkflow.model_validate('Sticking')
    )
    open_date: Optional[datetime] = Field(None, title='Open Date')
    close_date: Optional[datetime] = Field(None, title='Close Date')
    unavailable_hours: Optional[List[UnavailableHours]] = Field(
        [], title='Unavailable Hours'
    )
    counterparty_id: Optional[str] = Field(None, title='Counterparty Id')
    counterparty_name: Optional[str] = Field(None, title='Counterparty Name')
    in_network_terminals: Optional[List[InNetworkTerminal]] = Field(
        [], title='In Network Terminals'
    )
    in_network_supply_zones: Optional[List[InNetworkSupplyZone]] = Field(
        [], title='In Network Supply Zones'
    )
    supply_owner_id: Optional[str] = Field(None, title='Supply Owner Id')
    supply_owner_name: Optional[str] = Field(None, title='Supply Owner Name')
    freight_customer_id: Optional[str] = Field(None, title='Freight Customer Id')
    freight_customer_name: Optional[str] = Field(None, title='Freight Customer Name')
    freight_region_name: Optional[str] = Field(None, title='Freight Region Name')
    freight_region_id: Optional[str] = Field(None, title='Freight Region Id')
    layout_file: Optional[str] = Field(None, title='Layout File')
    layout_file_uploaded: Optional[datetime] = Field(None, title='Layout File Uploaded')
    auto_order_disabled: Optional[bool] = Field(False, title='Auto Order Disabled')
    updated_id: Optional[PydanticObjectId] = None
    trailer_config: Optional[str] = Field(None, title='Trailer Config')
    monitoring_strategy: Optional[MonitoringStrategy] = Field(
        default_factory=lambda: MonitoringStrategy.model_validate('Unmonitored')
    )
    estick_monitor_override: Optional[str] = Field(
        None, title='Estick Monitor Override'
    )
    notes: Optional[str] = Field(None, title='Notes')
    updated_on: Optional[datetime] = Field(None, title='Updated On')
    updated_by: Optional[str] = Field(None, title='Updated By')
    extra_data: Optional[Dict[str, Any]] = Field({}, title='Extra Data')
    credit_hold: Optional[bool] = Field(None, title='Credit Hold')
    compliance_hold_date: Optional[datetime] = Field(None, title='Compliance Hold Date')
    delivery_window_default_name: Optional[str] = Field(
        None, title='Delivery Window Default Name'
    )
    delivery_window_default_id: Optional[str] = Field(
        None, title='Delivery Window Default Id'
    )
    allow_by_product_request: Optional[bool] = Field(
        False, title='Allow By Product Request'
    )
    geofence: Optional[List[Coordinate]] = Field(None, title='Geofence')


class SupplyPriceUpdateManyRequest(BaseModel):
    id: Optional[str] = Field(None, title='Id')
    source_id: Optional[str] = Field(None, title='Source Id')
    source_system_id: Optional[str] = Field(None, title='Source System Id')
    contract: Optional[str] = Field(None, title='Contract')
    timezone: Optional[TimezoneEnum] = Field(
        default_factory=lambda: TimezoneEnum.model_validate('UTC')
    )
    effective_from: datetime = Field(..., title='Effective From')
    effective_to: datetime = Field(..., title='Effective To')
    price: float = Field(..., title='Price')
    price_type: PriceType
    terminal_id: Optional[constr(min_length=24, max_length=24)] = Field(
        None, title='Terminal Id'
    )
    terminal_source_id: Optional[str] = Field(None, title='Terminal Source Id')
    terminal_source_system_id: Optional[str] = Field(
        None, title='Terminal Source System Id'
    )
    terminal: Optional[str] = Field(None, title='Terminal')
    product_id: Optional[constr(min_length=24, max_length=24)] = Field(
        None, title='Product Id'
    )
    product_source_id: Optional[str] = Field(None, title='Product Source Id')
    product_source_system_id: Optional[str] = Field(
        None, title='Product Source System Id'
    )
    product: Optional[str] = Field(None, title='Product')
    supplier_id: Optional[constr(min_length=24, max_length=24)] = Field(
        None, title='Supplier Id'
    )
    supplier_source_id: Optional[str] = Field(None, title='Supplier Source Id')
    supplier_source_system_id: Optional[str] = Field(
        None, title='Supplier Source System Id'
    )
    supplier: Optional[str] = Field(None, title='Supplier')
    counterparty_id: Optional[constr(min_length=24, max_length=24)] = Field(
        None, title='Counterparty Id'
    )
    counterparty_source_id: Optional[str] = Field(None, title='Counterparty Source Id')
    counterparty_source_system_id: Optional[str] = Field(
        None, title='Counterparty Source System Id'
    )
    counterparty: Optional[str] = Field(None, title='Counterparty')
    store_id: Optional[constr(min_length=24, max_length=24)] = Field(
        None, title='Store Id'
    )
    store_source_id: Optional[str] = Field(None, title='Store Source Id')
    store_source_system_id: Optional[str] = Field(None, title='Store Source System Id')
    store_number: Optional[str] = Field(None, title='Store Number')
    enabled: Optional[bool] = Field(True, title='Enabled')
    disabled_until: Optional[datetime] = Field(None, title='Disabled Until')
    expire: Optional[datetime] = Field(None, title='Expire')
    min_quantity: Optional[int] = Field(None, title='Min Quantity')
    max_quantity: Optional[int] = Field(None, title='Max Quantity')
    curve_id: Optional[str] = Field(None, title='Curve Id')
    error: Optional[str] = Field(None, title='Error')
    row: Optional[int] = Field(None, title='Row')
    net_or_gross_type: Optional[BillableVolumeType] = None
    contract_lifting_valuation_method: Optional[BillableDateType] = None


class SurchargeAllReq(BaseModel):
    counterparty_name: Optional[str] = Field(None, title='Counterparty Name')
    counterparty_id: Optional[str] = Field(None, title='Counterparty Id')
    book_type: Optional[RateBookType] = None
    as_of: Optional[datetime] = Field(None, title='As Of')
    product_group: Optional[ProductGroups] = None
    freight_region_name: Optional[str] = Field(None, title='Freight Region Name')
    freight_region_id: Optional[str] = Field(None, title='Freight Region Id')


class SurchargeAllResp(BaseModel):
    id: PydanticObjectId
    counterparty: str = Field(..., title='Counterparty')
    counterparty_id: str = Field(..., title='Counterparty Id')
    product_group: Optional[ProductGroups] = None
    book_type: RateBookType
    freight_region_id: Optional[str] = Field(None, title='Freight Region Id')
    freight_region_name: Optional[str] = Field(None, title='Freight Region Name')
    type: SurchargeType
    effective_from: Optional[datetime] = Field(None, title='Effective From')
    effective_to: Optional[datetime] = Field(None, title='Effective To')
    surcharge: Optional[float] = Field(None, title='Surcharge')
    created_on: datetime = Field(..., title='Created On')
    created_by: Optional[str] = Field(None, title='Created By')


class SurchargeCreateReq(BaseModel):
    counterparty_name: Optional[str] = Field(None, title='Counterparty Name')
    counterparty_id: Optional[str] = Field(None, title='Counterparty Id')
    product_group: Optional[ProductGroups] = None
    book_type: RateBookType
    freight_region_id: Optional[str] = Field(None, title='Freight Region Id')
    freight_region_name: Optional[str] = Field(None, title='Freight Region Name')
    type: SurchargeType
    effective_from: datetime = Field(..., title='Effective From')
    effective_to: Optional[datetime] = Field(None, title='Effective To')
    surcharge: Optional[float] = Field(None, title='Surcharge')


class UpdateResponse(BaseModel):
    created: Optional[int] = Field(0, title='Created')
    end_dated: Optional[int] = Field(0, title='End Dated')
    bad_data: List[SupplyPriceUpdateResponse] = Field(..., title='Bad Data')
    duplicates: List[SupplyPriceUpdateResponse] = Field(..., title='Duplicates')
    exact_match: List[SupplyPriceUpdateResponse] = Field(..., title='Exact Match')


class UploadIMSRequest(BaseModel):
    store: str = Field(..., title='Store')
    tank: Union[str, int] = Field(..., title='Tank')
    timezone: Optional[TimezoneEnum] = None
    date: datetime = Field(..., title='Date')
    inches: Optional[int] = Field(None, title='Inches')
    volume: Optional[float] = Field(None, title='Volume')
    payload: Optional[Dict[str, Any]] = Field(None, title='Payload')


class UpsertLocationRequest(BaseModel):
    id: Optional[constr(min_length=24, max_length=24)] = Field(None, title='Id')
    source_id: Optional[str] = Field(None, title='Source Id')
    source_system_id: Optional[str] = Field(None, title='Source System Id')
    name: str = Field(..., title='Name')
    short_name: str = Field(..., title='Short Name')
    address: Optional[str] = Field(None, title='Address')
    type: TerminalType
    city: Optional[str] = Field(None, title='City')
    state: Optional[str] = Field(None, title='State')
    postal_code: Optional[str] = Field(None, title='Postal Code')
    market_id: Optional[str] = Field(None, title='Market Id')
    market: str = Field(..., title='Market')
    lat: float = Field(..., title='Lat')
    lon: float = Field(..., title='Lon')
    hours: Optional[str] = Field(None, title='Hours')
    contact: Optional[str] = Field(None, title='Contact')
    contact_phone: Optional[str] = Field(None, title='Contact Phone')
    active: Optional[bool] = Field(True, title='Active')
    include_in_backhaul: Optional[bool] = Field(False, title='Include In Backhaul')
    start_time: Optional[datetime] = Field(None, title='Start Time')
    end_time: Optional[datetime] = Field(None, title='End Time')
    requires_card: Optional[bool] = Field(False, title='Requires Card')
    splash_blending: Optional[bool] = Field(True, title='Splash Blending')
    split_load_terminals: Optional[List[str]] = Field([], title='Split Load Terminals')
    authorized_carriers: Optional[List[str]] = Field([], title='Authorized Carriers')
    extra_data: Optional[Dict[str, Any]] = Field(None, title='Extra Data')
    geofence: Optional[List[Union[Coordinate, List, str]]] = Field(
        None, title='Geofence'
    )
    supply_zones: Optional[List[str]] = Field([], title='Supply Zones')
    supply_map: Optional[List[SourceMap]] = Field(None, title='Supply Map')


class UpsertManyCounterpartyRequest(BaseModel):
    id: Optional[constr(min_length=24, max_length=24)] = Field(None, title='Id')
    source_id: Optional[str] = Field(None, title='Source Id')
    source_system_id: Optional[str] = Field(None, title='Source System Id')
    name: str = Field(..., title='Name')
    types: List[str] = Field(..., title='Types')
    carrier_type: Optional[str] = Field(None, title='Carrier Type')
    trailer_config: Optional[str] = Field(None, title='Trailer Config')
    emails: Optional[List[EmailStr]] = Field([], title='Emails')
    sourcing_strategy: Optional[SourcingStrategy] = None
    extra_data: Optional[Dict[str, Any]] = Field(None, title='Extra Data')
    goid: Optional[str] = Field(None, title='Goid')
    allow_short_loads: Optional[bool] = Field(False, title='Allow Short Loads')
    order_notification_preferences: Optional[List[OrderReqNotificationStates]] = Field(
        [], title='Order Notification Preferences'
    )
    supply_map: Optional[List[SourceMap]] = Field(None, title='Supply Map')
    available_credit: Optional[float] = Field(None, title='Available Credit')
    hold: Optional[bool] = Field(False, title='Hold')


class UpsertManyDriverRequest(BaseModel):
    model_config = ConfigDict(
        extra='allow',
    )
    id: Optional[constr(min_length=24, max_length=24)] = Field(None, title='Id')
    source_id: Optional[str] = Field(None, title='Source Id')
    source_system_id: Optional[str] = Field(None, title='Source System Id')
    name: str = Field(..., title='Name')
    username: str = Field(..., title='Username')
    email: Optional[EmailStr] = Field(None, title='Email')
    title: Optional[str] = Field(None, title='Title')
    hire_date: datetime = Field(..., title='Hire Date')
    termination_date: Optional[datetime] = Field(None, title='Termination Date')
    depot: Optional[str] = Field(None, title='Depot')
    depot_id: Optional[str] = Field(None, title='Depot Id')
    in_cab_trip_mode: Optional[InCabTripMode] = None
    in_cab_supply_option_mode: Optional[InCabSupplyOptionMode] = None
    preferred_template: Optional[str] = Field(None, title='Preferred Template')
    shift_preference: Shift
    shift_length_hours: Optional[float] = Field(None, title='Shift Length Hours')
    tractor_number: Optional[str] = Field(None, title='Tractor Number')
    tractor_id: Optional[str] = Field(None, title='Tractor Id')
    trailer_number: Optional[str] = Field(None, title='Trailer Number')
    trailer_id: Optional[str] = Field(None, title='Trailer Id')
    carded_terminals: Optional[List[str]] = Field(None, title='Carded Terminals')
    supervisor: Optional[str] = Field(None, title='Supervisor')
    active: Optional[bool] = Field(True, title='Active')


class UpsertManyStoreRequest(BaseModel):
    id: Optional[constr(min_length=24, max_length=24)] = Field(None, title='Id')
    store_number: str = Field(..., title='Store Number')
    name: Optional[str] = Field(None, title='Name')
    address: Optional[str] = Field(None, title='Address')
    city: Optional[str] = Field(None, title='City')
    state: Optional[str] = Field(None, title='State')
    postal_code: Optional[str] = Field(None, title='Postal Code')
    phone_number: Optional[str] = Field(None, title='Phone Number')
    emails: Optional[List[str]] = Field(None, title='Emails')
    hours: Optional[str] = Field(None, title='Hours')
    market: Optional[str] = Field(None, title='Market')
    market_id: Optional[str] = Field(None, title='Market Id')
    sector: Optional[str] = Field(None, title='Sector')
    sector_id: Optional[str] = Field(None, title='Sector Id')
    trailer_config: Optional[str] = Field(None, title='Trailer Config')
    lat: float = Field(..., title='Lat')
    lon: float = Field(..., title='Lon')
    timezone: Optional[TimezoneEnum] = Field(
        default_factory=lambda: TimezoneEnum.model_validate('America/Chicago')
    )
    status: Optional[StoreStatus] = Field(
        default_factory=lambda: StoreStatus.model_validate('open')
    )
    contact_name: Optional[str] = Field(None, title='Contact Name')
    open_date: Optional[datetime] = Field(None, title='Open Date')
    close_date: Optional[datetime] = Field(None, title='Close Date')
    unavailable_hours: Optional[List[UnavailableHoursRequest]] = Field(
        [], title='Unavailable Hours'
    )
    counterparty_id: Optional[str] = Field(None, title='Counterparty Id')
    counterparty_name: Optional[str] = Field(None, title='Counterparty Name')
    supply_owner_id: Optional[str] = Field(None, title='Supply Owner Id')
    supply_owner_name: Optional[str] = Field(None, title='Supply Owner Name')
    freight_customer_id: Optional[str] = Field(None, title='Freight Customer Id')
    freight_customer_name: Optional[str] = Field(None, title='Freight Customer Name')
    freight_region_name: Optional[str] = Field(None, title='Freight Region Name')
    freight_region_id: Optional[str] = Field(None, title='Freight Region Id')
    extra_data: Optional[Dict[str, Any]] = Field(None, title='Extra Data')
    monitoring_strategy: MonitoringStrategy
    estick_monitor_override: Optional[str] = Field(
        None, title='Estick Monitor Override'
    )
    drop_workflow: Optional[DropWorkflow] = Field(
        default_factory=lambda: DropWorkflow.model_validate('Sticking')
    )
    credit_hold: Optional[bool] = Field(None, title='Credit Hold')
    compliance_hold_date: Optional[datetime] = Field(None, title='Compliance Hold Date')
    allow_by_product_request: Optional[bool] = Field(
        None, title='Allow By Product Request'
    )
    in_network_terminals: Optional[List[str]] = Field(
        None, title='In Network Terminals'
    )
    delivery_window_default_name: Optional[str] = Field(
        None, title='Delivery Window Default Name'
    )
    delivery_window_default_id: Optional[str] = Field(
        None, title='Delivery Window Default Id'
    )
    geofence: Optional[List[Union[Coordinate, List, str]]] = Field(
        None, title='Geofence'
    )


class UpsertProductRequest(BaseModel):
    group: str = Field(..., title='Group')
    weight_group: Optional[str] = Field(None, title='Weight Group')
    extra_data: Optional[Dict[str, Any]] = Field(None, title='Extra Data')
    id: Optional[constr(min_length=24, max_length=24)] = Field(None, title='Id')
    source_id: Optional[str] = Field(None, title='Source Id')
    source_system_id: Optional[str] = Field(None, title='Source System Id')
    name: str = Field(..., title='Name')
    alias: Optional[str] = Field(None, title='Alias')
    enabled: Optional[bool] = Field(None, title='Enabled')
    specific_gravity: Optional[float] = Field(None, title='Specific Gravity')
    short_name: Optional[constr(max_length=10)] = Field(None, title='Short Name')
    icon: Optional[TankLidEnum] = None
    tank_lid_code_id: Optional[str] = Field(None, title='Tank Lid Code Id')
    supply_map: Optional[List[SourceMap]] = Field(None, title='Supply Map')


class UpsertTankRequest(BaseModel):
    id: Optional[str] = Field(None, title='Id')
    source_id: Optional[str] = Field(None, title='Source Id')
    source_system_id: Optional[str] = Field(None, title='Source System Id')
    store_id: Optional[constr(min_length=24, max_length=24)] = Field(
        None, title='Store Id'
    )
    tank_id: conint(gt=-1) = Field(..., title='Tank Id')
    island: Optional[str] = Field(None, title='Island')
    icon: Optional[TankLidEnum] = None
    product_id: Optional[str] = Field(None, title='Product Id')
    product: str = Field(..., title='Product')
    description: Optional[constr(max_length=1000)] = Field(None, title='Description')
    tank_chart: Optional[str] = Field(None, title='Tank Chart')
    storage_max: Optional[float] = Field(None, title='Storage Max')
    tank_size: Optional[float] = Field(None, title='Tank Size')
    fuel_bottom: Optional[float] = Field(None, title='Fuel Bottom')
    manufacturer: Optional[str] = Field(None, title='Manufacturer')
    dimensions: Optional[str] = Field(None, title='Dimensions')
    tank_color: Optional[str] = Field(None, title='Tank Color')
    minimum_load_size: Optional[int] = Field(None, title='Minimum Load Size')
    maximum_load_size: Optional[int] = Field(None, title='Maximum Load Size')
    strapping_table: Optional[str] = Field(None, title='Strapping Table')
    target_min: Optional[float] = Field(None, title='Target Min')
    target_max: Optional[float] = Field(None, title='Target Max')
    inventory_strategy: Optional[str] = Field(None, title='Inventory Strategy')
    load_tags: Optional[List[str]] = Field([], title='Load Tags')
    filter_tags: Optional[List[str]] = Field([], title='Filter Tags')
    split_me: Optional[bool] = Field(None, title='Split Me')
    water_reading_requirement: Optional[WaterReadingRequirement] = None
    requires_pump: Optional[bool] = Field(None, title='Requires Pump')
    carrier_id: Optional[str] = Field(None, title='Carrier Id')
    carrier: Optional[str] = Field(None, title='Carrier')
    extra_data: Optional[Dict[str, Any]] = Field({}, title='Extra Data')
    active: Optional[bool] = Field(True, title='Active')
    manifold_id: Optional[str] = Field(None, title='Manifold Id')
    apportion_percentage: Optional[float] = Field(None, title='Apportion Percentage')
    compliance_hold: Optional[bool] = Field(None, title='Compliance Hold')
    demand_model_tank_profile: Optional[str] = Field(
        None, title='Demand Model Tank Profile'
    )
    store_number: Optional[str] = Field(None, title='Store Number')



from typing import List
from fastapi import Query

from pydantic import BaseModel, EmailStr


class DomainInfo(BaseModel):
    domain: str
    role: str

class User(BaseModel):
    email: EmailStr
    password: str | None = None
    source: str = ""
    disabled: bool = False
    access: List[DomainInfo]
    password_reset_code: str | None = None


class Role(BaseModel):
    name: str
    scopes: List[str] = []


class RequireScope:
    def __init__(self, scope: str):
        self.scope = scope

    def __call__(self, scopes: str = Query(...)):
        return self.scope in scopes.split(",")



import json
from datetime import datetime

import pytz
from bb_integrations_lib.gravitate.rita_api import GravitateRitaAPI
from bb_integrations_lib.models.rita.config import MaxSync
from bb_integrations_lib.protocols.pipelines import Step
from loguru import logger
from typing import Any


class UpdateMaxSyncContextStep(Step):
    def __init__(self, rita_client: GravitateRitaAPI, config_id: str,
                 sync_date_tz: pytz.BaseTzInfo | None = None, *args, **kwargs):
        """
        Updates the max_sync context for a config, preserving the sync date.
        Used as a final step to persist pe_max_sync context after pipeline completion.

        :param rita_client: RITA API client
        :param config_id: The config ID to update max_sync for
        :param sync_date_tz: Timezone of MaxSyncDateTime in pe_max_sync. If provided,
            converts the date from this timezone to UTC.
        """
        super().__init__(*args, **kwargs)
        self.rita_client = rita_client
        self.config_id = config_id
        self.sync_date_tz = sync_date_tz

    def describe(self) -> str:
        return "Update Max Sync Context"

    async def execute(self, i: Any = None) -> Any:
        max_sync = self.pipeline_context.max_sync if self.pipeline_context else None
        if not max_sync:
            logger.warning("No max_sync available in pipeline context")
            return i

        # Get pe_max_sync from pipeline's extra_data (set by PEPriceExportStep)
        pe_max_sync = self.pipeline_context.extra_data.get('pe_max_sync')
        if not pe_max_sync:
            logger.warning("No pe_max_sync found in pipeline extra_data")
            return i

        updated_context = max_sync.context.copy() if max_sync.context else {}
        pe_max_sync_data = json.loads(pe_max_sync)

        if self.sync_date_tz and pe_max_sync_data.get('MaxSyncDateTime'):
            naive_dt = datetime.fromisoformat(pe_max_sync_data['MaxSyncDateTime'])
            if naive_dt.tzinfo is None:
                utc_dt = self.sync_date_tz.localize(naive_dt).astimezone(pytz.UTC)
                pe_max_sync_data['MaxSyncDateTime'] = utc_dt.isoformat()

        updated_context['pe_max_sync'] = pe_max_sync_data

        updated_max_sync = MaxSync(
            max_sync_date=max_sync.max_sync_date,
            context=updated_context
        )

        await self.rita_client.update_config_max_sync(
            config_id=self.config_id,
            max_sync=updated_max_sync
        )
        logger.success(f"Updated max_sync context for config {self.config_id}")

        return i



from datetime import datetime
from typing import Optional

from pydantic import BaseModel


class Bucket(BaseModel):
    """
    Bucket is a grouping mechanism. It may reference another bucket as its container
    Items may reference a bucket as their container, but an item can only be contained in one bucket.
    """
    owned_by: Optional[str] = None # TODO: Write a validator that checks for cycles
    name: str
    description: Optional[str] = None
    updated_by: Optional[str] = None
    updated_on: Optional[datetime] = None
    is_active: bool = True



from io import BytesIO
from loguru import logger

import pandas as pd

from bb_integrations_lib.gravitate.sd_api import GravitateSDAPI
from bb_integrations_lib.protocols.pipelines import Step, ParserBase, Input, Output
from bb_integrations_lib.shared.model import RawData
from pandas import DataFrame
from datetime import datetime, date, UTC


class BBDExportPayrollStep(Step):
    def __init__(self, sd_client: GravitateSDAPI, file_name: str, parser: type[ParserBase] | None = None,
                 parser_kwargs: dict | None = None, export_file: bool = True, target_date: str | None = None,
                 additional_ep_params: dict | None = None, *args, **kwargs):
        """
        Export a payroll file from S&D.

        :param export_file: If True, call the endpoint to export the file as an XLSX. If False, use the endpoint that
          provides JSON.
        """
        super().__init__(*args, **kwargs)
        self.sd_client = sd_client
        self.file_name = file_name
        self.export_file_only = export_file
        self.additional_endpoint_parameters = additional_ep_params or {
            "status": None,
            "driver_id": None,
            "updated_after": None,
        }
        if parser:
            self.custom_parser = parser
            self.custom_parser_kwargs = parser_kwargs or {}
        self.target_date = target_date

    def describe(self):
        return "Export Payroll from Supply and Dispatch"

    async def execute(self, last_sync_date: datetime | None = None) -> RawData:
        last_sync_date = self.last_sync_date(last_sync_date)
        dt = datetime.now(UTC)
        if self.export_file_only is True:
            logger.info(f"Exporting payroll file only for {last_sync_date}")
            df = await self.export_payroll_file_only(last_sync_date)
        else:
            logger.info(f"Exporting payroll json for {last_sync_date}")
            df =  await self.export_json(last_sync_date)
        return RawData(data=df.to_csv().encode("utf-8"), file_name=f"{self.file_name}_{dt.strftime("%Y%m%d%H%M%S")}")

    def last_sync_date(self, last_sync_date: datetime | None = None) -> str:
        if last_sync_date is None:
            if self.target_date is not None:
                last_sync_date = self.target_date
            elif self.pipeline_context.max_sync is not None:
                last_sync_date = self.pipeline_context.max_sync.max_sync_date.isoformat()
            else:
                last_sync_date = datetime.combine(date.today(), datetime.min.time()).isoformat()
        else:
            last_sync_date = last_sync_date.isoformat()
        return last_sync_date

    async def export_payroll_file_only(self, dt: datetime) -> DataFrame:
        resp = await self.sd_client.payroll_export_file(date=dt)
        df = pd.read_excel(BytesIO(resp.content),
                           dtype={"driver_source_id": str},
                           engine="openpyxl",
                           keep_default_na=False)
        if df.empty:
            raise Exception("No payroll data found")
        if hasattr(self, "custom_parser"):
            parser = self.custom_parser(tenant_name=self.tenant_name, **self.custom_parser_kwargs)
            df = await parser.parse(df)
            parser_logs = parser.get_logs()
            self.pipeline_context.included_files["parser_logs"] = parser_logs
        return df

    async def export_json(self, dt: datetime) -> DataFrame:
        resp = await self.sd_client.payroll_export(date=dt, **self.additional_endpoint_parameters)
        json_resp = resp.json()
        df = self.pre_parse_json(json_resp)
        if df.empty:
            raise Exception("No payroll data found")
        return df

    def pre_parse_json(self, json_resp: dict) -> DataFrame:
        all_rows = []
        for payroll_record in json_resp:
            header_info = {k: v for k, v in payroll_record.items() if k != 'detail'}
            details = payroll_record.get('detail', [])
            if not details:
                all_rows.append(header_info)
            else:
                for detail in details:
                    row = header_info.copy()
                    row.update(detail)
                    all_rows.append(row)
        df = pd.DataFrame(all_rows)
        datetime_columns = [
            'start_date', 'end_date', 'updated', 'shift_start',
            'shift_actual_start', 'shift_actual_end', 'overridden_datetime'
        ]
        for col in datetime_columns:
            if col in df.columns:
                df[col] = pd.to_datetime(df[col], errors='coerce')

        return df


import ftplib
from abc import ABC
from io import BytesIO, StringIO
from pathlib import Path
from stat import S_ISREG
from typing import Iterable

import paramiko
import tenacity
from loguru import logger
from paramiko import RSAKey
from tenacity import retry_if_exception_type, stop_after_attempt, wait_fixed

from bb_integrations_lib.provider.ftp.model import FTPFileInfo, FTPAuthType, FTPType
from bb_integrations_lib.secrets.credential_models import FTPCredential
from bb_integrations_lib.shared.model import RawData, File


class FTPClientInterface(ABC):
    def __init__(self, credentials: FTPCredential):
        self.credentials = credentials
        self.first_connection = True

    def connect(self):
        self.first_connection = False

    def disconnect(self):
        raise NotImplementedError()

    def _reconnect(self, retry_state: tenacity.RetryCallState):
        raise NotImplementedError()

    def cwd(self, directory: str) -> None:
        raise NotImplementedError()

    def list_files(self, directory: str) -> Iterable[str]:
        raise NotImplementedError()

    def rename_files(self, files: Iterable[tuple[str, str]]) -> None:
        raise NotImplementedError()

    def delete_files(self, paths: Iterable[str]) -> None:
        raise NotImplementedError()

    def upload_files(self, files: Iterable[File], path: str) -> None:
        raise NotImplementedError()

    def download_files(self, paths: Iterable[str]) -> Iterable[RawData]:
        raise NotImplementedError()

    @staticmethod
    def reconnect_retry(func):
        def wrap(*args, **kwargs):
            if args[0].first_connection:
                args[0].connect()
            r = tenacity.Retrying(
                reraise=True,
                retry=retry_if_exception_type((OSError, AttributeError)),
                stop=stop_after_attempt(3),
                wait=wait_fixed(3),
                after=args[0]._reconnect
            )
            for attempt in r:
                with attempt:
                    return func(*args, **kwargs)

        return wrap

class FTPClient(FTPClientInterface):
    def __init__(self, credentials: FTPCredential):
        super().__init__(credentials)

        if self.credentials.ftp_type not in [FTPType.ftp, FTPType.ftps, FTPType.ftpes]:
            raise NotImplementedError(
                f"Attempted to use FTPClient with unsupported FTP type: {self.credentials.ftp_type} "
                "(only supports ftp, ftps, ftpes)"
            )

        self.is_tls = self.credentials.ftp_type in [FTPType.ftps, FTPType.ftpes]
        self.client: ftplib.FTP | None = None

    def connect(self):
        if self.is_tls:
            self.client = ftplib.FTP_TLS(
                host=self.credentials.host,
                user=self.credentials.username,
                passwd=self.credentials.password,
            )
            if self.credentials.ftp_type == FTPType.ftpes:
                self.client.prot_p()
        else:
            self.client = ftplib.FTP(
                host=self.credentials.host,
                user=self.credentials.username,
                passwd=self.credentials.password
            )
        super().connect()

    def disconnect(self):
        self.client.quit()
        # Will force a reconnection next time it is used
        self.first_connection = True

    def _reconnect(self, retry_state: tenacity.RetryCallState):
        # Targeted by reconnect_retry
        self.connect()

    @FTPClientInterface.reconnect_retry
    def list_files(self, directory: str) -> Iterable[str]:
        # Paths might be absolute - for consistency with the SFTP client, make them relative.
        return [Path(x).name for x in self.client.nlst(directory)]

    @FTPClientInterface.reconnect_retry
    def rename_files(self, files: Iterable[tuple[str, str]]) -> None:
        for old_name, new_name in files:
            logger.debug(f"Renaming file {old_name} -> {new_name}")
            self.client.rename(old_name, new_name)

    @FTPClientInterface.reconnect_retry
    def delete_files(self, paths: Iterable[str]) -> None:
        for path in paths:
            logger.debug(f"Deleting file {path}")
            self.client.delete(path)

    @FTPClientInterface.reconnect_retry
    def upload_files(self, files: Iterable[File], path: str) -> None:
        for file in files:
            name = file.file_name
            logger.debug(f"Uploading file {name}")
            self.client.storbinary(f"STOR {path}/{name}", File.to_bytes(file.file_data))


    @FTPClientInterface.reconnect_retry
    def download_files(self, paths: Iterable[str]) -> Iterable[RawData]:
        results = []
        for path in paths:
            logger.debug(f"Downloading {path}")
            buf = BytesIO()
            self.client.retrbinary(f"RETR {path}", buf.write)
            buf.seek(0)
            p = Path(path)
            results.append(RawData(file_name=p.name, data=buf))
        return results

    @FTPClientInterface.reconnect_retry
    def get_file_info(self, path: str) -> FTPFileInfo:
        # Size is guaranteed
        size = self.client.size(path)

        return FTPFileInfo(
            size=size,
            permissions=None,
            owner_id=None,
            group_id=None,
            last_access_time=None,
            last_modification_time=None,
        )


class SFTPClient(FTPClientInterface):
    def __init__(self, credentials: FTPCredential, private_key_path: str | None = None):
        super().__init__(credentials)
        self.ssh = paramiko.SSHClient()
        self.ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        self.sftp: paramiko.SFTPClient | None = None
        self.private_key_path = private_key_path

    def connect(self):
        if self.credentials.private_key:
            pkey = RSAKey.from_private_key(
                StringIO(self.credentials.private_key),
                self.credentials.passphrase
            )
        elif self.credentials.auth_type == FTPAuthType.rsa:
            pkey = get_ppk()
        else:
            pkey = None


        self.ssh.connect(
            hostname=self.credentials.host,
            username=self.credentials.username,
            password=self.credentials.password,
            port=self.credentials.port or 22,
            key_filename=self.private_key_path,
            pkey=pkey,
            passphrase=self.credentials.passphrase,
            look_for_keys=False,
            allow_agent=False,
        )
        self.sftp = self.ssh.open_sftp()
        super().connect()

    def disconnect(self):
        self.sftp.close()
        self.first_connection = True

    def _reconnect(self, retry_state: tenacity.RetryCallState):
        self.connect()

    @FTPClientInterface.reconnect_retry
    def cwd(self, directory: str) -> None:
        self.sftp.chdir(directory)

    @FTPClientInterface.reconnect_retry
    def list_files(self, directory: str) -> Iterable[str]:
        logger.debug(f"Listing {directory}")
        res = self.sftp.listdir_attr(directory)
        return [x.filename for x in res if S_ISREG(x.st_mode)]


    @FTPClientInterface.reconnect_retry
    def rename_files(self, files: Iterable[tuple[str, str]]) -> None:
        for old_path, new_path in files:
            logger.debug(f"Renaming {old_path} -> {new_path}")
            return self.sftp.rename(old_path, new_path)

    @FTPClientInterface.reconnect_retry
    def delete_files(self, paths: Iterable[str]) -> None:
        for path in paths:
            logger.debug(f"Deleting {path}")
            self.sftp.remove(path)

    @FTPClientInterface.reconnect_retry
    def upload_files(self, files: Iterable[File], path: str) -> None:
        for file in files:
            name = file.file_name
            logger.debug(f"Uploading file {name}")
            self.sftp.putfo(file.to_bytes(file.file_data), f"{path}/{name}")

    @FTPClientInterface.reconnect_retry
    def download_files(self, paths: Iterable[str]) -> Iterable[RawData]:
        results = []
        for path in paths:
            logger.debug(f"Downloading {path}")
            buf = BytesIO()
            self.sftp.getfo(path, buf)
            buf.seek(0)
            p = Path(path)
            results.append(RawData(file_name=p.name, data=buf))
        return results

    @FTPClientInterface.reconnect_retry
    def get_file_info(self, path: str) -> FTPFileInfo:
        file_info = self.sftp.stat(path)
        return FTPFileInfo(
            size=file_info.st_size,
            permissions=oct(file_info.st_mode) if file_info.st_mode else None,
            owner_id=file_info.st_uid,
            group_id=file_info.st_gid,
            last_access_time=file_info.st_atime,
            last_modification_time=file_info.st_mtime,
        )

def get_ppk(path: str = 'secrets/id_rsa') -> RSAKey:
    """
        Given a path, try to load RSA key file
        :param path: The path to load
        :return: The loaded RSA key
    """
    try:
        private_key_path = Path(path)
        if not private_key_path.exists():
            raise FileNotFoundError(f"Private key file not found at: {private_key_path}")
        private_key: RSAKey = RSAKey.from_private_key_file(str(private_key_path))
        return private_key
    except FileNotFoundError as fne:
        msg = f'Failed to load private key from file: {path} -> {fne}'
        logger.error(msg)
        raise
    except Exception as e:
        msg = f'Failed to load private key from file: {path} -> {e}'
        logger.error(msg)
        raise


import os
import uuid
from typing import Any

import loguru

from bb_integrations_lib.protocols.pipelines import Step
from bb_integrations_lib.provider.ftp.client import FTPIntegrationClient
from bb_integrations_lib.shared.model import FileConfigRawData, RawData


class ArchiveSFTPStep(Step):
    def __init__(self, ftp_client: FTPIntegrationClient, src_directory: str | None = None,
                 archive_directory: str | None = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.ftp_client = ftp_client
        self.src_directory = src_directory
        self.archive_directory = archive_directory

    def describe(self) -> str:
        return "SFTP Rename Step"

    async def execute(self, i: Any) -> Any:
        if isinstance(i, FileConfigRawData):
            old_name = os.path.join(i.file_config.inbound_directory, i.file_name)
            new_name = os.path.join(i.file_config.archive_directory, i.file_name)
        elif isinstance(i, RawData):
            if self.src_directory is None or self.archive_directory is None:
                raise RuntimeError("Attempted to archive with a RawData object but src_directory or archive_directory was not provided.")
            old_name = os.path.join(self.src_directory, i.file_name)
            new_name = os.path.join(self.archive_directory, i.file_name)
        else:
            raise NotImplementedError(f"Unsupported input type: {type(i)}")
        try:
            loguru.logger.debug(f"Archiving file {old_name} -> {new_name}")
            self.ftp_client.rename_file(old_name, new_name)
        except Exception as e:
            try:
                loguru.logger.debug(f"Archiving file failed...")
                # this file may already exist. Give the file a randomized name and then archive.
                new_name = new_name.replace(i.file_name, f"DUPLICATE_{uuid.uuid4()}_{i.file_name}")
                loguru.logger.debug(f"Archiving file (backup attempt) {old_name} -> {new_name}")
                self.ftp_client.rename_file(old_name, new_name)
            except:
                loguru.logger.warning(f"Archiving backup file failed. Deleting source file to prevent duplicate readings.")
                self.ftp_client.delete_file(old_name)
        loguru.logger.debug(f"Archived file {old_name} -> {new_name}")
        return i


import json

from loguru import logger

from bb_integrations_lib.gravitate.sd_api import GravitateSDAPI
from bb_integrations_lib.protocols.pipelines import Step
from bb_integrations_lib.shared.model import SDSetOrderExportStatusRequest, ERPStatus, SDGetUnexportedOrdersResponse


class BBDFinalizeExportStatuses(Step):
    """
    Final step for order export pipelines. Sets export statuses for all orders
    based on pipeline results. Reads all data from pipeline_context.extra_data:
        - unexported_orders: original orders list
        - errored_orders: orders that errored during processing
        - processed_orders: orders that were successfully processed
        - rollback_processed: if True, rolls back processed orders to 'pending'

    On rollback, processed orders are set to 'pending'. Otherwise, they are set to 'sent'.
    Any orders not accounted for in errors or processed are marked as 'errors'.
    """

    def __init__(self,
                 sd_client: GravitateSDAPI,
                 step_key: str | None = "Order Export Results",
                 *args,
                 **kwargs):
        super().__init__(*args, **kwargs)
        self.sd_client = sd_client
        self.step_key = step_key

    def describe(self):
        return "Collect all results and set order export status"

    async def execute(self) -> None:
        orders: list[SDGetUnexportedOrdersResponse] = self.pipeline_context.extra_data.get("unexported_orders", [])
        errors = self.pipeline_context.extra_data.get("errored_orders", [])
        processed = self.pipeline_context.extra_data.get("processed_orders", [])
        rollback_processed = self.pipeline_context.extra_data.get("rollback_processed", False)

        all_requests: list[SDSetOrderExportStatusRequest] = []

        errored_orders_requests = format_status_requests(errors, ERPStatus.errors)
        all_requests.extend(errored_orders_requests)

        if rollback_processed:
            rollback_order_requests = format_status_requests(processed, ERPStatus.pending)
            all_requests.extend(rollback_order_requests)
            logger.warning("Error while exporting. Rolling back order status to 'pending'.")
            rollback_requests = json.dumps([order.model_dump() for order in rollback_order_requests], indent=2)
            self.pipeline_context.included_files[f"{self.step_key} - rolled back orders"] = rollback_requests
        else:
            processed_orders_requests = format_status_requests(processed, ERPStatus.sent)
            all_requests.extend(processed_orders_requests)
            processed_json = json.dumps([order.model_dump() for order in processed_orders_requests], indent=2)
            self.pipeline_context.included_files[f"{self.step_key} - processed orders"] = processed_json


        # Set filtered orders back to pending
        filtered = self.pipeline_context.extra_data.get("filtered_orders", [])
        if filtered:
            filtered_requests = [
                SDSetOrderExportStatusRequest(order_id=str(order.order_id), status=ERPStatus.pending)
                for order in filtered
            ]
            all_requests.extend(filtered_requests)

        # Catch any orders that were in the input but never ended up in errors or processed
        accounted_ids = {str(order.get("_id") or order.get("order_id")) for order in errors + processed}
        missing_order_ids = [ str(order.order_id) for order in orders if str(order.order_id) not in accounted_ids]
        missing_orders_requests = []
        if missing_order_ids:
            logger.error(f"These orders were missed by the pipeline: {missing_order_ids}")
            missing_orders_requests = [
                SDSetOrderExportStatusRequest(
                    order_id=str(order.order_id),
                    status=ERPStatus.errors,
                    error="Missed by the pipeline" #TODO: should this be errors
                )
                for order in orders
                if str(order.order_id) not in accounted_ids
            ]
            all_requests.extend(missing_orders_requests)

        errored_json = json.dumps([order.model_dump() for order in errored_orders_requests + missing_orders_requests], indent=2)
        self.pipeline_context.included_files[f"{self.step_key} - errored orders"] = errored_json

        # Send all requests at once
        response = await self.sd_client.bulk_set_export_order_status(all_requests)
        response.raise_for_status()

def format_status_requests(orders: list[dict], status: ERPStatus) -> list[SDSetOrderExportStatusRequest]:
    status_requests = [
        SDSetOrderExportStatusRequest(
            order_id=str(order.get("_id") or order.get("order_id")),
            status=status,
            error=order.get("error_message") if status == ERPStatus.errors else None,
        )
        for order in orders
    ]

    return status_requests



from datetime import datetime, UTC
from enum import Enum
from typing import Optional, Dict, Self
import json
import pandas as pd
from functools import cached_property
from bson.objectid import ObjectId
from pydantic import BaseModel, field_validator, Field


class MappingType(str, Enum):
    site = "site"
    tank = "tank"
    counterparty = "counterparty"
    credential = "credential"
    product = "product"
    terminal = "terminal"
    driver = "driver"
    depot = "depot"
    trailer = "trailer"
    tractor = "tractor"
    other = "other"
    composite = "composite"


class MapType(str, Enum):
    parent = "parent"
    child = "child"


class View(str, Enum):
    grid = "grid"
    gallery = "gallery"


class CompositeMapKey(BaseModel):
    """Multi-field composite key for mapping lookups.

    A composite key allows mappings to be identified by multiple fields
    instead of a single source_id. For example, a mapping might be uniquely
    identified by (tenant, product, terminal) rather than just source_id.

    Example:
        >>> key = CompositeMapKey(key={"tenant": "acme", "product": "fuel"})
        >>> key.to_cache_key()
        '{"product":"fuel","tenant":"acme"}'
        >>> key.matches({"tenant": "acme"})
        True
    """
    model_config = {"frozen": True}

    key: dict[str, str] = Field(..., description="Field-value pairs forming the composite key")

    @field_validator('key', mode='before')
    @classmethod
    def validate_key(cls, v: dict) -> dict[str, str]:
        if not v:
            raise ValueError("Composite key cannot be empty")
        validated = {}
        for k, val in v.items():
            if not isinstance(k, str) or not k.strip():
                raise ValueError(f"Key must be non-empty string, got: {k}")
            if not isinstance(val, str):
                raise ValueError(f"Value for '{k}' must be string, got {type(val).__name__}")
            validated[k.strip()] = val
        return validated

    @cached_property
    def _cache_key_str(self) -> str:
        """Cached canonical string representation for hashing."""
        return json.dumps(self.key, sort_keys=True, separators=(',', ':'))

    def to_cache_key(self) -> str:
        """Generate a canonical string representation for caching/hashing (JSON format)."""
        return self._cache_key_str

    @classmethod
    def from_string(cls, s: str) -> Self:
        """Parse a composite key from its JSON string representation."""
        if not s:
            raise ValueError("Cannot parse empty string")
        try:
            key_dict = json.loads(s)
            if not isinstance(key_dict, dict):
                raise ValueError("Invalid format: expected JSON object")
            return cls(key=key_dict)
        except json.JSONDecodeError as e:
            raise ValueError(f"Failed to parse composite key: {e}")

    @cached_property
    def _cached_hash(self) -> int:
        return hash(self._cache_key_str)

    def __hash__(self) -> int:
        return self._cached_hash

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, CompositeMapKey):
            return False
        return self.key == other.key

    def __str__(self) -> str:
        return self.to_cache_key()

    def __repr__(self) -> str:
        return f"CompositeMapKey({self.to_cache_key()})"

    def __lt__(self, other: 'CompositeMapKey') -> bool:
        if not isinstance(other, CompositeMapKey):
            return NotImplemented
        return self.to_cache_key() < other.to_cache_key()

    def matches(self, partial: dict[str, str]) -> bool:
        """
        Check if this key contains all fields from the partial key.

        Returns True if all key-value pairs in partial exist in this key.
        Short-circuits on first mismatch for performance.
        """
        if len(partial) > len(self.key):
            return False
        for k, v in partial.items():
            if self.key.get(k) != v:
                return False
        return True

    def get(self, field: str, default: str | None = None) -> str | None:
        """Get a specific component from the composite key."""
        return self.key.get(field, default)

    @property
    def fields(self) -> list[str]:
        """Return the list of field names in this composite key."""
        return list(self.key.keys())


class MapBase(BaseModel):
    updated_by: str | None = Field(default="admin", description="User who last updated the mapping")
    updated_on: datetime | None = Field(default_factory=lambda: datetime.now(UTC), description="Timestamp of last update")
    type: Optional[MappingType | None] = Field(default=None, description="Type of mapping (site, tank, counterparty, etc.)")
    source_system: str | None = Field(default=None, description="External system the mapping originates from")
    is_active: Optional[bool] = Field(default=True, description="Whether the mapping is currently active")
    source_name: str | None = Field(default=None, description="Display name in the source system")
    gravitate_name: str | None = Field(default=None, description="Display name in Gravitate")
    extra_data: Optional[Dict] = Field(default=None, description="Additional metadata for the mapping")

    @field_validator('type', 'source_system', mode='before')
    @classmethod
    def validate_type_source_system(cls, v):
        if isinstance(v, str) and (v == 'nan' or v == ''):
            return None
        elif isinstance(v, float) and pd.isna(v):  # Check for actual NaN (float nan)
            return None
        return v


class Children(MapBase):
    source_id: str | None = Field(
        default=None,
        description="Source system identifier"
    )
    gravitate_id: str | None = Field(
        default=None,
        description="Gravitate identifier"
    )
    id: Optional[str] = Field(
        default_factory=lambda: str(ObjectId()),
        description="Unique identifier for this child mapping"
    )

    @field_validator('source_id', 'gravitate_id', mode='before')
    @classmethod
    def validate_identifiers(cls, v):
        if v is None or v == '' or v == 'nan':
            return None
        if isinstance(v, float) and pd.isna(v):
            return None
        return str(v)



class Map(MapBase):
    id: Optional[str] = Field(
        default=None,
        description="Mapping document ID"
    )
    source_id: str | CompositeMapKey = Field(
        ...,
        description="Source system identifier (string or composite key)"
    )
    gravitate_id: str | CompositeMapKey = Field(
        ...,
        description="Gravitate identifier (string or composite key)"
    )
    children: Optional[list[Children]] = Field(
        default_factory=list,
        description="List of child mappings"
    )
    children_type: MappingType | None = Field(
        default=None,
        description="Type of child mappings"
    )
    owning_bucket_id: Optional[str] = Field(
        default=None,
        description="Bucket that owns this mapping"
    )
    group_id: str | None = Field(
        default=None,
        description="Group that owns this mapping"
    )


    def is_composite(self) -> bool:
        """Check if this mapping uses composite keys."""
        return isinstance(self.source_id, CompositeMapKey)

    def source_id_str(self) -> str | None:
        """Get source_id as string (for simple IDs) or None (for composite)."""
        return self.source_id if isinstance(self.source_id, str) else None

    def source_id_composite(self) -> CompositeMapKey | None:
        """Get source_id as CompositeMapKey or None (for simple IDs)."""
        return self.source_id if isinstance(self.source_id, CompositeMapKey) else None

    def gravitate_id_str(self) -> str | None:
        """Get gravitate_id as string (for simple IDs) or None (for composite)."""
        return self.gravitate_id if isinstance(self.gravitate_id, str) else None

    def gravitate_id_composite(self) -> CompositeMapKey | None:
        """Get gravitate_id as CompositeMapKey or None (for simple IDs)."""
        return self.gravitate_id if isinstance(self.gravitate_id, CompositeMapKey) else None


class UpdateMapItem(Map):
    id: str
    view: Optional[str] = None
    bucket_id: Optional[str] = None


WORKER_BACKGROUND_REQUEST = "worker.background.request"
"""The Kafka topic that is used to send requests to background workers."""
WORKER_BACKGROUND_RESPONSE = "worker.background.response"
"""The Kafka topic that background workers will use to reply with status updates."""
WORKER_IMMEDIATE_REQUEST = "worker.immediate.request"
"""The Kafka topic that immediate (semi-synchronous / RPC) worker requests are sent to."""
WORKER_IMMEDIATE_RESPONSE = "worker.immediate.response"
"""The Kafka topic that immediate (semi-synchronous / RPC) worker responses are sent to."""
WORKER_CROSSROADS_REQUEST = "worker.crossroads.request"
"""The Kafka topic that crossroads worker requests are sent to."""
WORKER_CROSSROADS_RESPONSE = "worker.crossroads.response"
"""The Kafka topic that crossroads workers will send responses to."""


def build_topic(namespace: str, topic: str) -> str:
    """
    Builds a Kafka topic string in a consistent way using a namespace (usually rita or rita-test) and a topic name.
    See the constants in bb_integrations_lib.workers.topics for topic names.
    """
    return f"{namespace}.{topic}"
